Package: FastKRR
Type: Package
Title: Kernel Ridge Regression using 'RcppArmadillo'
Version: 0.2.1
Authors@R: 
  c(person(given = "Gyeongmin",
           family = "Kim",
           role = c("aut"),
           email = "rlarudals0824@gmail.com",
           comment = "Sungshin Women's University"),
    person(given = "Seyoung",
           family = "Lee",
           role = c("aut"),
           email = "sudang0404@gmail.com",
           comment = "Sungshin Women's University"),
    person(given = "Miyoung",
           family = "Jang",
           role = c("aut"),
           email = "miyoung9072@gmail.com",
           comment = "Sungshin Women's University"),
    person(given = "Kwan-Young",
           family = "Bak",
           role = c("aut", "cre", "cph"),
           email = "kybak@sungshin.ac.kr",
           comment = c(ORCID = "0000-0002-4541-160X", "Sungshin Women's University")))
Description: Provides core computational operations in C++ via 'RcppArmadillo', enabling faster performance than pure R, improved numerical stability, and parallel execution with OpenMP where available. On systems without OpenMP support, the package automatically falls back to single-threaded execution with no user configuration required. For efficient model selection, it integrates with 'CVST' to provide sequential-testing cross-validation and additionally supports restricted maximum likelihood (REML) for continuous optimization of the regularization parameter. The package offers a unified interface for exact kernel ridge regression and three scalable approximations—Nyström, Pivoted Cholesky, and Random Fourier Features—allowing analyses with substantially larger sample sizes than are feasible with exact KRR. It also integrates with the 'tidymodels' ecosystem via the 'parsnip' model specification 'krr_reg', and the S3 method tunable.krr_reg(). To understand the theoretical background, one can refer to Wainwright (2019) <doi:10.1017/9781108627771>.
License: GPL (>= 2)
URL: https://github.com/kybak90/FastKRR, https://www.tidymodels.org
BugReports: https://github.com/kybak90/FastKRR/issues
Imports: CVST, generics, parsnip, Rcpp, rlang, tibble, ggplot2
LinkingTo: Rcpp, RcppArmadillo
Suggests: knitr, rmarkdown, dials, tidymodels, modeldata, dplyr,
        testthat (>= 3.0.0)
SystemRequirements: OpenMP (optional)
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-07-21 09:10:23 UTC; KGM
Author: Gyeongmin Kim [aut] (Sungshin Women's University),
  Seyoung Lee [aut] (Sungshin Women's University),
  Miyoung Jang [aut] (Sungshin Women's University),
  Kwan-Young Bak [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-4541-160X>, Sungshin Women's
    University)
Maintainer: Kwan-Young Bak <kybak@sungshin.ac.kr>
Repository: CRAN
Date/Publication: 2026-07-21 09:30:18 UTC

# Generated by roxygen2: do not edit by hand

S3method(coef,krr)
S3method(error,default)
S3method(error,krr)
S3method(param,default)
S3method(param,krr)
S3method(plot,krr)
S3method(predict,krr)
S3method(print,approx_kernel)
S3method(print,kernel_matrix)
S3method(print,krr)
S3method(summary,krr)
S3method(tunable,krr_reg)
S3method(update,krr_reg)
export(approx_kernel)
export(error)
export(fastkrr)
export(krr_reg)
export(make_kernel)
export(param)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(generics,tunable)
importFrom(ggplot2,aes)
importFrom(ggplot2,element_text)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_minimal)
importFrom(stats,formula)
importFrom(stats,kernel)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.response)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,rcauchy)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,time)
importFrom(stats,update)
importFrom(tibble,tibble)
useDynLib(FastKRR, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# FastKRR

<!-- badges: start -->
<!-- badges: end -->

The ‘FastKRR’ implements its core computational operations in C++ via
‘RcppArmadillo’, enabling faster performance than pure R, improved
numerical stability, and parallel execution with OpenMP where available.
On systems without OpenMP support, the package automatically falls back
to single-threaded execution with no user configuration required. For
efficient model selection, it integrates with ‘CVST’ to provide full and
sequential-testing cross-validation and additionally supports restricted
maximum likelihood (REML) for continuous optimization of the
regularization parameter. The package offers a unified interface for
exact kernel ridge regression and three scalable approximations—Nyström,
Pivoted Cholesky, and Random Fourier Features—allowing analyses with
substantially larger sample sizes than are feasible with exact KRR. It
also integrates with the ‘tidymodels’ ecosystem via the ‘parsnip’ model
specification ‘krr_reg’, the S3 method ‘tunable.krr_reg()’, and the
direct fitting helper ‘fit_krr()’.

**Dependencies:** Rcpp, RcppArmadillo, CVST, parsnip  
This package uses **CVST** (GPL ≥ 2). Overall license: **GPL (≥ 2)**.

- Authors:
  - Gyeongmin Kim, Sungshin Women’s University,
    <rlarudals0824@gmail.com>
  - Seyoung Lee, Sungshin Women’s University, <sudang0404@gmail.com>
  - Miyoung Jang, Sungshin Women’s University, <miyoung9072@gmail.com>
  - Kwan-Young Bak, professor at Sungshin Women’s University,
    <kybak@sungshin.ac.kr>,
    [ORCID:0000-0002-4541-160X](https://orcid.org/0000-0002-4541-160X%7D%7BORCID:0000-0002-4541-160X)

## Installation

FastKRR is available both on CRAN (stable release) and GitHub
(development version).

### Option 1: Install from CRAN

This installs the latest stable version. On macOS (with Clang compiler),
CRAN binaries are built without OpenMP support. If you need OpenMP, see
the note below about installing from source.

``` r
# CRAN binary (no OpenMP on macOS)
install.packages("FastKRR")            

# CRAN source build (enables OpenMP if configured)
# install.packages("FastKRR", type="source")  
```

### Option 2: Install from GitHub

This installs the development version (always built from source).

``` r
# install.packages("pak")
pak::pak("kybak90/FastKRR")
```

<!-- ## Parallelization -->
<!-- Some functions in **FastKRR** support parallel computation via OpenMP. -->
<!-- - On Windows and most Linux systems, OpenMP is available by default, and computations will use multiple threads. -->
<!-- - On macOS, if OpenMP is not installed, the package will run in single-threaded mode automatically. -->
<!-- No special installation steps are required. -->

## Example

This is a basic example of fitting a Gaussian kernel ridge regression
estimator to a dataset ${(x_i, y_i)}_{i=1}^n$ using (1) exact
computation, (2) pivoted Cholesky decomposition, (3) Nyström
approximation, and (4) random Fourier features.

``` r
library(FastKRR)

# example data set
set.seed(1)
n = 1000
rho = 1
X = runif(n, 0, 1)
y = sin(2*pi*X^3) + rnorm(n, 0, 0.1)

data = data.frame(X, y = y)

# model fitting - exact
model_exact = fastkrr(data = data, response = "y", kernel = "gaussian", rho = rho, opt = "exact", verbose = FALSE)

# model fitting - pivoted
model_pivoted = fastkrr(data = data, response = "y", kernel = "gaussian", rho = rho, opt = "pivoted", verbose = FALSE)

# model fitting - nystrom
model_nystrom = fastkrr(data = data, response = "y", kernel = "gaussian", rho = rho, opt = "nystrom", verbose = FALSE)

# model fitting - rff
model_rff = fastkrr(data = data, response = "y", kernel = "gaussian", rho = rho, opt = "rff", verbose = FALSE)


# prediction
new_n = 500
new_x = matrix(runif(new_n, 0, 1), nrow = new_n)

pred_exact = predict(model_exact, new_x)
pred_pivoted = predict(model_pivoted, new_x)
pred_nystrom = predict(model_nystrom, new_x)
pred_rff = predict(model_rff, new_x)
```

The visualization of the fitted results is shown below.

``` r
library(ggplot2)
datas = data.frame(new_x)
datas$pred_exact = as.numeric(pred_exact)
datas$pred_pivoted = as.numeric(pred_pivoted)
datas$pred_nystrom = as.numeric(pred_nystrom)
datas$pred_rff = as.numeric(pred_rff)


ggplot(data = datas) +
  geom_line(aes("x" = new_x, "y" = pred_exact, color = 'opt = "exact"'), linewidth=1.2) + 
  geom_line(aes("x" = new_x, "y" = pred_pivoted, color = 'opt = "pivoted"'), linewidth=1.2) + 
  geom_line(aes("x" = new_x, "y" = pred_nystrom, color = 'opt = "nystrom"'), linewidth=1.2) + 
  geom_line(aes("x" = new_x, "y" = pred_rff, color = 'opt = "rff"'), linewidth=1.2) +
  geom_point(data = data, aes(x = X, y = y), col = "gray", alpha = 0.2) +
  theme_minimal()
```

<img src="man/figures/README-unnamed-chunk-4-1.png" width="100%" />

## OpenMP support on macOS with Clang compiler

### 1) Install OpenMP runtime (Homebrew)

    # Ensure Command Line Tools (if not already installed)
    xcode-select --install || true

    # Install/update Homebrew libomp
    brew update
    brew install libomp

### 2) Configure `~/.R/Makevars` (works for Apple Silicon & Intel)

Copy the block below into `~/.R/Makevars` (create the file if it does
not exist).

    # OpenMP on macOS with Apple Clang 
    # Try to detect Homebrew prefix; fallback covers both Apple Silicon & Intel.
    HOMEBREW_PREFIX := $(shell brew --prefix 2>/dev/null)
    ifeq ($(strip $(HOMEBREW_PREFIX)),)
      HOMEBREW_PREFIX := /opt/homebrew
    endif

    # Include & lib paths (also add Intel path as extra fallback)
    CPPFLAGS  += -I$(HOMEBREW_PREFIX)/opt/libomp/include -I/usr/local/opt/libomp/include
    LDFLAGS   += -L$(HOMEBREW_PREFIX)/opt/libomp/lib     -L/usr/local/opt/libomp/lib

    # Standard R variables for OpenMP
    SHLIB_OPENMP_CFLAGS   = -Xpreprocessor -fopenmp
    SHLIB_OPENMP_CXXFLAGS = -Xpreprocessor -fopenmp
    SHLIB_OPENMP_FCFLAGS  = -fopenmp
    SHLIB_OPENMP_FFLAGS   = -fopenmp
    SHLIB_OPENMP_LIBS     = -lomp

> Restart R after editing `~/.R/Makevars` so the flags are picked up.

### 3) Install FastKRR

By default, CRAN binary packages on macOS are built without OpenMP,
because Apple’s toolchain does not enable it when CRAN compiles
binaries. To enable OpenMP, you must install from source after setting
up your system:

``` r
# Install from CRAN source
install.packages("FastKRR", type = "source")

# Or install directly from GitHub (also source build)
pak::pak("kybak90/FastKRR")
```

### 4) Example & OpenMP confirmation (inside FastKRR)

``` r
# example data set
set.seed(1)
n = 1000
rho = 1
X = runif(n, 0, 1)
y = sin(2*pi*X^3) + rnorm(n, 0, 0.1)

data = data.frame(X, y = y)

# model fitting - nystrom
model_nystrom = fastkrr(data = data, response = "y", kernel = "gaussian", rho = rho, opt = "nystrom", verbose = FALSE)

model_nystrom$n_threads    # >1 indicates OpenMP used by FastKRR (default 4)
#> [1] 4
```

### 5) Show that FastKRR links to libomp (macOS only)

The following code chunk inspects the dynamic libraries linked to the
compiled FastKRR shared object (`.so` or `.dylib`).

- If the output of `otool -L` contains a line with **`libomp.dylib`**,  
  then the package binary is linked against the OpenMP runtime and  
  **parallelization via OpenMP is available**.  
- If `libomp.dylib` is absent, the package was compiled without OpenMP
  support.

``` r
is_macos = (Sys.info()[["sysname"]] == "Darwin")

if (is_macos) {
  so_dir = system.file("libs", .Platform$r_arch, package = "FastKRR")
  if (so_dir == "") so_dir = system.file("libs", package = "FastKRR")
  so_files = list.files(so_dir, pattern = "\\.(so|dylib)$", full.names = TRUE)

  so = if (length(so_files)) so_files[[1]] else ""
  cat("FastKRR shared object:", so, "\n")

  if (nzchar(so)) {
    cmd = paste("otool -L", shQuote(so))
    cat(system(cmd, intern = TRUE), sep = "\n")
  } else {
    cat("No shared object found. Is FastKRR installed with compiled code?\n")
  }
} else {
  cat("Skipping otool check (not macOS).\n")
}
```

When you run this on macOS, a successful OpenMP build will show a line
such as:

    /Library/Frameworks/R.framework/Versions/4.3-arm64/Resources/lib/libomp.dylib

This confirms that OpenMP is enabled for FastKRR on your system.

#' Extract Model Coefficients from a Fitted KRR Model
#'
#' @description
#' Extracts the estimated coefficients from a fitted Kernel Ridge Regression (KRR) model.
#' The type of coefficient reported depends on the kernel approximation method:
#' for \code{opt = "exact"}, \code{"nystrom"}, or \code{"pivoted"}, the coefficients represent
#' the dual weights \eqn{\alpha}. For \code{opt = "rff"}, they represent the coefficients \eqn{\beta}.
#'
#' @param object An S3 object of class \code{krr}, typically returned by
#'   \code{\link{fastkrr}}.
#' @param ... Additional arguments (currently ignored).
#'
#' @return A numeric vector of the estimated model coefficients (\eqn{\alpha} or \eqn{\beta}).
#'
#' @seealso \code{\link{fastkrr}}
#'
#' @examples
#' # Data setting
#' set.seed(1)
#' lambda = 1e-4
#' d = 1
#' n = 50
#' rho = 1
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
#' y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)
#'
#' data = data.frame(X, y = y)
#'
#' # Example: exact
#' model = fastkrr(data = data, response = "y",
#'                  kernel = "gaussian", opt = "exact",
#'                  rho = rho, lambda = lambda)
#'
#' coef(model)
#'
#' @rdname coef.krr
#' @method coef krr
#' @export
coef.krr = function(object, ...){
  return(as.vector(object$coefficients))
}

#' Summary method for fitted Kernel Ridge Regression models
#'
#' @description
#' Computes and displays a comprehensive summary of a fitted Kernel Ridge Regression (KRR) model,
#' including the original function call, fitted hyperparameters (via \code{\link{param.krr}}),
#' and the final training mean squared error (via \code{\link{error.krr}}).
#'
#' @param object An S3 object of class \code{krr}, typically returned by
#'   \code{\link{fastkrr}}.
#' @param ... Additional arguments (currently ignored).
#'
#' @return Invisibly returns the hyperparameter list or model statistics after
#'   printing the summary panel to the console.
#'
#' @seealso \code{\link{fastkrr}}, \code{\link{param.krr}}, \code{\link{error.krr}}
#'
#' @examples
#' # Data setting
#' set.seed(1)
#' lambda = 1e-4
#' d = 1
#' n = 50
#' rho = 1
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
#' y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)
#'
#' data = data.frame(X, y = y)
#'
#' # Example: exact
#' model = fastkrr(data = data, response = "y",
#'                  kernel = "gaussian", opt = "exact",
#'                  rho = rho, lambda = lambda)
#'
#' summary(model)
#'
#' @export
summary.krr = function(object, ...){
  model = object

  invisible(param(model))

  cat("Training error:\n")
  print(error(model))
}

#' Compute Model Error for Kernel Ridge Regression Models
#'
#' @description
#' Computes the model error for kernel ridge regression (\code{"krr"} object).
#' Returns the mean squared error (MSE) between the observed responses
#' and the fitted values stored in the object.
#'
#' @param x An object of class \code{"krr"}, typically returned by
#'   \code{\link{fastkrr}}.
#' @param data_new An optional data frame containing new predictor variables
#'   and the observed response variable. The response column must have the
#'   same name as the response variable used to fit the model. If
#'   \code{NULL}, the training MSE is computed. Otherwise, the prediction MSE (PMSE)
#'   is computed using predictions for \code{data_new}.
#' @param ... Additional arguments (ignored).
#'
#' @return A numeric value giving the mean squared error (MSE). If \code{data_new = NULL},
#'   this returns the training MSE based on the fitted values. If \code{data_new} is provided,
#'   it returns the prediction mean squared error (PMSE) evaluated on the new dataset.
#'
#' @details
#' This method computes the mean squared error defined as:
#' \deqn{\text{MSE} = \frac{1}{n} \sum_{i = 1} (y_i - \hat{y}_i)^2}
#' Depending on the presence of \code{data_new}, the components are defined as follows:
#' \itemize{
#'   \item If \code{data_new = NULL}, \eqn{y} is the observed response vector and
#'     \eqn{\hat{y}} is the fitted values vector, both stored within the \code{"krr"} object.
#'   \item If \code{data_new} is provided, \eqn{y} is the observed response extracted from
#'     \code{data_new} and \eqn{\hat{y}} is the predicted values vector generated by
#'     applying \code{\link{predict.krr}} to the new predictor variables.
#' }
#'
#' @seealso \code{\link{summary.krr}}, \code{\link{plot.krr}}, \code{\link{predict.krr}}
#'
#'
#' @examples
#' # Data setting
#' set.seed(1)
#' lambda = 1e-4
#' d = 1
#' n = 50
#' rho = 1
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
#' y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)
#'
#' data = data.frame(X, y = y)
#'
#' model = fastkrr(data = data, response = "y", kernel = "gaussian",
#'                  opt = "exact", lambda = lambda)
#'
#' # MSE
#' error(model)
#'
#' new_n = 50
#' new_x = matrix(runif(new_n*d, 0, 1), nrow = new_n, ncol = d)
#' new_y = as.vector(sin(2*pi*rowMeans(new_x)^3) + rnorm(new_n, 0, 0.1))
#' new_data = data.frame(new_x, y = new_y)
#'
#' # PMSE
#' error(model, data_new = new_data)
#'
#' @name error.krr
NULL


#' @rdname error.krr
#' @export
error = function(x, ...) {
  UseMethod("error")
}

#' @rdname error.krr
#' @export
error.krr = function(x, data_new = NULL, ...) {
  model = x

  if (is.null(data_new)){ # Training MSE
    return(as.numeric(
      crossprod(model$y - model$fitted.values) / length(model$y)
    ))
  }else{ # Prediction MSE
    if (!is.data.frame(data_new)) stop("data_new must be a data.frame")
    if (anyNA(data_new)) stop("data_new contains missing values (NA).")
    if (!model$call$response %in% names(data_new)) stop("Response variable not found in data_new.")

    y_new = data_new[[model$call$response]]
    x_new = as.matrix(
      data_new[, setdiff(names(data_new), model$call$response), drop = FALSE]
    )

    pred = predict(model, newdata = x_new)

    as.numeric(
      crossprod(y_new - pred) / length(y_new)
    )
  }
}

test_that("error.krr computes training MSE correctly when data_new is NULL", {
  # Set seed for reproducibility
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  # Fit a FastKRR model
  model <- fastkrr(data = df, response = "y", kernel = "gaussian",
                   opt = "exact", rho = 1.0, lambda = 1e-3)

  # Compute training MSE
  train_mse <- error(model)

  # Manual expectation check
  expected_mse <- mean((df$y - model$fitted.values)^2)

  # Validate output type, structure, and numeric correctness
  expect_type(train_mse, "double")
  expect_length(train_mse, 1)
  expect_equal(train_mse, expected_mse)
})

test_that("error.krr computes prediction MSE (PMSE) correctly when data_new is provided", {
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  model <- fastkrr(data = df, response = "y", kernel = "gaussian",
                   opt = "exact", rho = 1.0, lambda = 1e-3)

  # Construct new test data frame
  new_n <- 20
  new_X <- matrix(runif(new_n * d), nrow = new_n, ncol = d)
  colnames(new_X) <- c("x1", "x2")
  new_y <- sin(new_X[, 1]) + rnorm(new_n, sd = 0.1)
  new_df <- data.frame(new_X, y = new_y)

  # Compute prediction MSE (PMSE)
  pmse <- error(model, data_new = new_df)

  # Manual prediction expectation check
  preds <- predict(model, newdata = as.matrix(new_X))
  expected_pmse <- mean((new_y - preds)^2)

  expect_type(pmse, "double")
  expect_length(pmse, 1)
  expect_equal(pmse, expected_pmse)
})

test_that("error.krr throws expected errors on invalid data_new inputs", {
  set.seed(123)
  n <- 20
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  model <- fastkrr(data = df, response = "y", kernel = "gaussian",
                   opt = "exact", rho = 1.0, lambda = 1e-3)

  # 1. Error when data_new is not a data.frame
  expect_error(
    error(model, data_new = matrix(1:10, nrow = 5)),
    "data_new must be a data.frame"
  )

  # 2. Error when data_new contains NA values
  na_df <- data.frame(x1 = c(1, NA), x2 = c(2, 3), y = c(1, 2))
  expect_error(
    error(model, data_new = na_df),
    "data_new contains missing values"
  )

  # 3. Error when the response column is missing in data_new
  wrong_col_df <- data.frame(x1 = runif(5), x2 = runif(5), target = runif(5))
  expect_error(
    error(model, data_new = wrong_col_df),
    "Response variable not found in data_new"
  )
})

test_that("summary.krr prints model summary and returns training error", {
  # Set up reproducible synthetic data
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  # Fit a FastKRR model
  fit <- fastkrr(
    data = df,
    response = "y",
    kernel = "gaussian",
    opt = "exact",
    rho = 1.0,
    lambda = 1e-3
  )

  # Verify console output sections from param() and error() calls
  expect_output(summary(fit), "Call:")
  expect_output(summary(fit), "Parameters:")
  expect_output(summary(fit), "Training error:")

  # Capture the return value of summary() which returns training error
  res_err <- summary(fit)

  # Validate that the returned value is a positive numeric training error
  expect_type(res_err, "double")
  expect_true(is.numeric(res_err))
  expect_gt(res_err, 0)
  expect_equal(res_err, as.numeric(error(fit)))
})

#' Predict responses for new data using fitted KRR model
#'
#' Generates predictions from a fitted Kernel Ridge Regression (KRR) model
#' for new data.
#'
#' @param object A S3 object of class \code{krr} created by \code{\link{fastkrr}}.
#' @param newdata A numeric design matrix containing new observations
#'                for which predictions are to be made. If \code{newdata} is missing, the function returns fitted values.
#' @param ... Additional arguments (currently ignored).
#'
#'
#' @return A numeric vector of predicted values corresponding to \code{newdata} or fitted values.
#'
#' @seealso \code{\link{fastkrr}}, \code{\link{make_kernel}}
#'
#' @examples
#' # Data setting
#' n = 30
#' d = 1
#'
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
#' y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)
#'
#' data = data.frame(X, y = y)
#'
#' lambda = 1e-4
#' rho = 1
#'
#' # Fitting model: pivoted
#' model = fastkrr(data = data, response = "y",
#'                 kernel = "gaussian", rho = rho, lambda = lambda, opt = "pivoted")
#'
#' # Predict
#' new_n = 50
#' new_x = matrix(runif(new_n*d, 0, 1), nrow = new_n, ncol = d)
#' new_y = as.vector(sin(2*pi*rowMeans(new_x)^3) + rnorm(new_n, 0, 0.1))
#'
#' pred = predict(model, new_x)
#' crossprod(pred - new_y) / new_n
#'
#' @importFrom stats predict
#' @export
predict.krr = function(object, newdata, ...){
  if(missing(newdata)) return(object$fitted.values)

  if(object$opt == "rff"){
    x_new = newdata
    W = object$W
    b = object$b
    coef = object$coefficients
    Z_new = make_Z(x_new, W = W, b = b)

    return(as.vector(Z_new %*% coef))
  }else{
    x_new = newdata
    x = object$x
    kernel = object$kernel
    rho = object$rho
    coef = object$coefficients

    K_new = make_kernel(x, x_new, kernel = kernel, rho = rho)

    return(as.vector(K_new %*% coef))
  }
}


#' Fit kernel ridge regression using exact or approximate methods
#'
#' This function performs kernel ridge regression (KRR).
#' The regularization parameter \eqn{\lambda} can be supplied by
#' the user or selected automatically using cross-validation or
#' restricted maximum likelihood (REML). For scalability,
#' three different kernel approximation strategies are supported (Nyström approximation,
#' Pivoted Cholesky decomposition, Random Fourier Features(RFF)), and kernel matrix
#' can be computed using two methods (Gaussian kernel, Laplace kernel).
#'
#' @param data A data frame containing the data point variables and response
#' variable.
#' @param response A character string specifying the name of the response
#' variable in \code{data}.
#' @param kernel Kernel type either "gaussian"or "laplace".
#' @param rho Scaling parameter of the kernel(\eqn{\rho}),  specified by the user.
#'   Defaults to \code{1}.
#' \deqn{\text{Gaussian kernel : } \mathcal{K}(x, x') = \exp(-\rho \| x - x'\|^2_2)}
#' \deqn{\text{Laplace kernel : } \mathcal{K}(x, x') = \exp(-\rho \| x - x'\|_1)}
#' @param m Approximation rank(number of random features) used for the low-rank kernel approximation.
#'   If not provided by the user, it defaults to \eqn{\lceil n^{1/2} \cdot \log(d + 5) \rceil},
#'   where \eqn{n = nrow(X)} and \eqn{d = ncol(X)}. Also it must be a positive integer.
#' @param eps Tolerance parameter used only in \code{"pivoted"}
#'   for the stopping criterion of the Pivoted Cholesky decomposition.
#'   The default value is dynamically adjusted based on the mathematical smoothness
#'   of the selected kernel: it defaults to \code{1e-6} for the infinitely
#'   differentiable and smooth Gaussian kernel, and relaxes to \code{1e-4} for
#'   the non-smooth, rougher Laplace kernel to properly balance numerical precision
#'   and computational overhead.
#' @param lambda Regularization parameter. If \code{NULL}, the penalty parameter
#'   is chosen automatically via \pkg{CVST} package or REML: for \code{selection_method = "REML"}
#'   it defaults to the search range \eqn{[10^{-11}, 10^{-1}]} (a length-2 min/max vector), and for
#'   \code{selection_method \%in\% c("exactCV", "fastCV")} it defaults to a grid of 100 values over
#'   \eqn{[10^{-11}, 10^{-1}]}.
#' @param opt Method for constructing or approximating :
#'  \describe{
#'   \item{\code{"exact"}}{Construct the full kernel matrix
#'   \eqn{K \in \mathbb{R}^{n\times n}} using design matrix \eqn{X}.}
#'   \item{\code{"nystrom"}}{Construct a low-rank approximation of
#'       the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
#'       using the Nyström approximation.}
#'   \item{\code{"pivoted"}}{Construct a low-rank approximation of
#'       the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
#'       using Pivoted Cholesky decomposition.}
#' \item{\code{"rff"}}{Use Random Fourier Features to construct a feature map
#'   \eqn{Z \in \mathbb{R}^{n \times m}} (with \eqn{m} random features) so that
#'   \eqn{K \approx Z Z^\top}. Here, \eqn{m} is the number of features.}

#'  }
#' @param n_threads Number of parallel threads. If \code{NULL}, it defaults to
#'   half of the available system processors (\code{max_threads \%/\% 2}).
#'   Note that parallelization (implemented in C++) is applied for \code{opt = "nystrom"},
#'   \code{opt = "rff"}, \code{kernel = "laplace"}, or \code{selection_method = "REML"}.
#'   For these parallelizable cases, if the system has 3 or fewer processors, it automatically
#'   falls back to \code{1} thread; otherwise, it is capped at \code{max_threads - 1}.
#'   For all other settings, it is restricted to \code{1} thread.
#' @param selection_method Method used to select \eqn{\lambda} when a grid or \code{NULL} is
#'   passed. One of:
#'   \describe{
#'     \item{\code{"exactCV"}}{Full cross-validation via \pkg{CVST} (default).}
#'     \item{\code{"fastCV"}}{Accelerated sequential-testing CV via \pkg{CVST}.}
#'     \item{\code{"REML"}}{Restricted Maximum Likelihood.}
#'   }
#' @param verbose If TRUE, detailed progress and cross-validation
#' results are printed to the console. If FALSE, suppresses
#' intermediate output and only returns the final result.
#' @param na.rm Logical. If \code{TRUE}, rows containing missing values are
#'   removed before fitting. Defaults to \code{FALSE}.
#'
#' @details
#' The function performs several input checks and automatic adjustments:
#'
#' \itemize{
#'   \item \code{lambda} can be specified in four ways:
#'     \enumerate{
#'       \item A positive numeric scalar, in which case the model is fitted with
#'          this single value and no selection is performed (any \code{selection_method}).
#'       \item A numeric vector of length 2 giving \code{c(min, max)}; only valid when
#'          \code{selection_method = "REML"}, which optimizes \eqn{\lambda} within this range.
#'       \item A numeric vector (length >= 3) of positive values used as a tuning grid;
#'          only valid when \code{selection_method \%in\% c("exactCV", "fastCV")}, and
#'          selection is performed by \pkg{CVST} cross-validation (sequential testing if
#'          \code{selection_method = "fastCV"}).
#'       \item \code{NULL}: use a default range/grid (internal setting) and tune \code{lambda}
#'         via \pkg{CVST} or REML, depending on \code{selection_method}.}
#'}
#'
#' After fitting the model with \code{fastkrr()}, users can readily generate predictions
#' for new test data or out-of-sample observations using the standard generic
#' \code{\link[stats]{predict}} function, which is internally dispatched to \code{\link{predict.krr}}.
#'
#'
#' @return
#'
#' \itemize{
#'   \item{\code{coefficients}: Estimated coefficient vector. Accessible via \code{model$coefficients}.
#'     For \code{opt = "rff"}, this is an \eqn{m}-dimensional weight vector in the random feature space.
#'     For all other options (\code{"exact"}, \code{"pivoted"}, \code{"nystrom"}),
#'     this is an \eqn{n}-dimensional dual coefficient vector.}
#'   \item{\code{fitted.values}: Fitted values \eqn{\mathbb{R}^{n}}. Accessible via \code{model$fitted.values}.
#'     Computed as:
#'     \itemize{
#'       \item \eqn{\hat{y} = K \hat{\alpha}} for \code{opt = "exact"} (full kernel matrix \eqn{K}).
#'       \item \eqn{\hat{y} = \tilde{K} \tilde{\alpha}} for \code{opt = "pivoted"} or \code{"nystrom"} (low-rank kernel matrix \eqn{\tilde{K}}).
#'       \item \eqn{\hat{y} = Z \hat{\beta}} for \code{opt = "rff"} (random feature matrix \eqn{Z}).
#'     }
#'     Here, \eqn{\hat{\alpha}}, \eqn{\tilde{\alpha}}, and \eqn{\hat{\beta}} represent the estimated regression coefficient vectors (\code{coefficients}) obtained under each respective option \code{opt}.}
#'   \item{\code{opt}: Kernel approximation option.
#'   One of \code{"exact"}, \code{"pivoted"}, \code{"nystrom"}, \code{"rff"}.}
#'   \item{\code{kernel}: Kernel used (\code{"gaussian"} or \code{"laplace"}).}
#'   \item{\code{x}: Input design matrix.}
#'   \item{\code{y}: Response vector.}
#'   \item{\code{lambda}: Regularization parameter. If \code{NULL}, tuned
#'   by cross-validation via \pkg{CVST} or REML.}
#'   \item{\code{rho}: Additional user-specified hyperparameter.}
#'   \item{\code{selection_method}: Tunning method for select hyperparmeter lambda}
#'   \item{\code{call}: The matched function call used to create the object.}
#'   \item{\code{n_threads}: Number of threads used for parallelization.}
#'   \item{\code{removed_row_idx}: Indices of rows removed due to missing values.}
#' }
#'
#' Additional components depend on the value of \code{opt}:
#'
#' \subsection{opt = \dQuote{exact}}{
#' \itemize{
#'   \item{\code{K}: The full kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}.}
#'   \item{\code{chol_factor}: Lower triangular Cholesky factor \eqn{L \in \mathbb{R}^{n \times n}}
#'     of \eqn{K + n\lambda I}; satisfies \eqn{K + n\lambda I = L L^\top}.}
#' }}
#'
#'
#' \subsection{opt = \dQuote{nystrom}}{
#' \itemize{
#'   \item{\code{m}: Kernel approximation degree.}
#'   \item{\code{approx_factor}: The method provides a low-rank approximation to the kernel matrix
#'     \eqn{R \in \mathbb{R}^{n \times m}} obtained via
#'     Nyström approximation; satisfies \eqn{K \approx R R^\top}.}
#' }}
#'
#' \subsection{opt = \dQuote{pivoted}}{
#' \itemize{
#'   \item{\code{m}: Kernel pproximation degree.}
#'   \item{\code{approx_factor}: The method provides a low-rank approximation to the kernel matrix
#'     \eqn{PR \in \mathbb{R}^{ n \times m}} obtained via
#'     Pivoted Cholesky decomposition; satisfies \eqn{K \approx PR\,(PR)^\top}.}
#'   \item{\code{eps}: Numerical tolerance used for early stopping in the pivoted Cholesky decomposition.}
#' }}
#'
#' \subsection{opt = \dQuote{rff}}{
#' \itemize{
#'   \item{\code{m}: Number of random features.}
#'   \item{\code{approx_factor}: Random Fourier Feature matrix \eqn{Z \in \mathbb{R}^{n \times m}} with
#'     \eqn{Z_{ij} = z_j(x_i) = \sqrt{2/m}\cos(\omega_j^\top x_i + b_j), \quad j = 1, \cdots, m,}
#'     so that \eqn{K \approx Z Z^\top}.}
#'   \item{\code{W}: Random frequency matrix \eqn{\omega \in \mathbb{R}^{m \times d}}
#'       (row \eqn{j} is \eqn{\omega_j^\top \in \mathbb{R}^d}), drawn i.i.d. from the spectral density of the chosen kernel:
#'       \itemize{
#'         \item {Gaussian: \eqn{\omega_{jk} \sim \mathcal{N}(0, 2\gamma)} (e.g., \eqn{\gamma=1/\ell^2}).}
#'         \item {Laplace: \eqn{\omega_{jk} \sim \mathrm{Cauchy}(0, 1/\sigma)} i.i.d.
#'       }}}
#'   \item{\code{b} Random phase vector \eqn{b \in \mathbb{R}^m}, i.i.d. \eqn{\mathrm{Unif}[0,\,2\pi]}.}
#' }}
#'
#' @seealso \code{\link{predict.krr}}, \code{\link{param.krr}}, \code{\link{make_kernel}}
#'
#'
#' @examples
#' # Data setting
#' set.seed(1)
#' lambda = 1e-4
#' d = 3
#' rho = 1
#' n = 50
#'
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
#' y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)
#'
#' data = data.frame(X, y = y)
#'
#' # Example: pivoted cholesky
#' model = fastkrr(data = data, response = "y",  kernel = "gaussian",
#'                 opt = "pivoted", rho = rho, lambda = lambda)
#'
#' # Example: nystrom
#' model = fastkrr(data = data, response = "y", kernel = "gaussian",
#'                 opt = "nystrom", rho = rho, lambda = lambda)
#'
#' # Example: random fourier features
#' model = fastkrr(data = data, response = "y", kernel = "gaussian",
#'                 opt = "rff", rho = rho, lambda = lambda)
#'
#' # Example: Laplace kernel
#' model = fastkrr(data = data, response = "y",  kernel = "laplace",
#'                 opt = "nystrom", n_threads = 1, rho = rho)
#'
#' # Generate predictions for new data
#' new_X = matrix(runif(10 * d, 0, 1), nrow = 10, ncol = d)
#' pred = predict(model, new_X)
#' print(pred)
#'
#' @export
fastkrr = function(data, response,
                   kernel = "gaussian", # c(gaussian, laplace)
                   opt = "exact",  # c(exact, pivoted, nystrom, rff)
                   m = NULL,
                   eps = NULL,
                   rho = 1,
                   lambda = NULL,
                   selection_method = "exactCV",  # c(exactCV, fastCV, REML)
                   n_threads = NULL,
                   verbose =  TRUE,
                   na.rm = FALSE)
{
  call = match.call()

  # Data frame
  if(!is.data.frame(data)) stop("data must be a data.frame", call. = FALSE)
  removed_row_idx = integer(0)
  if (anyNA(data)){
    na_cols = names(data)[colSums(is.na(data)) > 0]
    if(!na.rm){
      stop("Missing values (NA) found in: ",
           paste(na_cols, collapse = ", "),
           "\nSet na.rm = TRUE to remove rows containing missing values.",
           call. = FALSE)
    }else{
      removed_row_idx = which(!stats::complete.cases(data))
      data = stats::na.omit(data)
      message("Removed ", length(removed_row_idx),
              " row", if (length(removed_row_idx) != 1) "s" else "",
              " containing missing values.\n",
              "Removed row indices: ",
              paste(removed_row_idx, collapse = ", "), ".")
    }
  }
  if(!(length(response) == 1 && is.character(response) &&
       response %in% colnames(data)))
    stop("response must be a single column name in data.", call. = FALSE)

  # Kernel and approximation option
  if (!kernel %in% c("gaussian", "laplace"))
    stop("kernel must be one of 'gaussian', 'laplace'", call. = FALSE)
  if (!opt %in% c("exact", "pivoted", "nystrom", "rff"))
    stop("opt must be one of 'exact', 'pivoted', 'nystrom', 'rff'", call. = FALSE)
  if (!selection_method %in% c("exactCV", "fastCV", "REML"))
    stop("selection_method must be one of 'exactCV', 'fastCV', 'REML'", call. = FALSE)

  y = data[[response]]
  x = as.matrix(data[, setdiff(names(data), response), drop = FALSE])

  # Parameter setting
  if (is.null(eps)){
    eps = if (kernel == "gaussian") 1e-6 else 1e-4
  }else{
    if (eps <= 0) stop("eps must be a positive real number", call. = FALSE)}
  if (is.null(m))
    m = as.integer(max(1, ceiling(nrow(x)^(1/2) * log(ncol(x) + 5))))
  else if(m <= 0)
    stop("m must be a positive integer", call. = FALSE)
  if (rho <= 0)
    stop("rho must be a positive real number", call. = FALSE)


  # Complexity parameter setting
  if(is.null(lambda)){
    if(selection_method == "REML"){
      lambda = c(1e-11, 1e-1) # default min, max for REML
    }else{
      lambda = 10^seq(-11, -1, len = 100) # default grid for exactCV/fastCV
    }
  }else if(is.numeric(lambda) && length(lambda) == 1 && lambda > 0){
    lambda = lambda # scalar: lambda fixed, selection not performed
  }else if(is.vector(lambda) && all(lambda > 0) && length(lambda) == 2){
    if(selection_method != "REML")
      stop("A lambda vector of length 2 (min, max) is only supported when selection_method = 'REML'.",
           call. = FALSE)
    lambda = lambda # vector: REML min, max
  }else if(is.vector(lambda) && all(lambda > 0) && length(lambda) >= 3){
    if(selection_method == "REML")
      stop("REML does not support a lambda grid of length >= 3; provide a length-2 vector (min, max), a single value, or NULL.",
           call. = FALSE)
    lambda = lambda # vector: CV grid
  }else{
    stop("lambda must be a positive number, or a numeric vector of positive real numbers of length 2 (REML min/max) or length >= 3 (CV grid)",
         call. = FALSE)
  }


  # Adjust number of threads
  max_threads = get_num_procs()
  if (is.null(n_threads)) n_threads = as.integer(max(1, max_threads %/% 2))
  if ((opt %in% c("nystrom", "rff")) || (kernel == "laplace") || (selection_method == "REML")){
    n_threads = if (max_threads <= 3) 1 else min(n_threads, max_threads - 1)
  }else{
    n_threads = 1
  }

  n = nrow(x)
  d = ncol(x)
  m = as.integer(m)
  #rate = round(m / n, 5)
  idx_ny = NULL  # will be set during REML nystrom selection; reused in fitting

  # Lambda selection
  if(length(lambda) > 1){
    if(selection_method == "REML"){
      lambda = if(opt == "exact"){
        K_tmp = make_kernel(x, kernel = kernel, rho = rho, n_threads = n_threads)
        reml_exact(K_tmp, y, lambda, n_threads = n_threads)
      }else if(opt == "pivoted"){
        PR_tmp = pchol_kernel(x, rho = rho, kernel = kernel, m = m, eps = eps, verbose = FALSE)$PR
        reml_lowrank(PR_tmp, y, lambda, n_threads = n_threads)
      }else if(opt == "nystrom"){
        idx_ny = sample(seq_len(n), m)
        K_nm_tmp = make_kernel(x[idx_ny, , drop = FALSE], x,
                               kernel = kernel, rho = rho, n_threads = n_threads)
        K_mm_tmp = make_kernel(x[idx_ny, , drop = FALSE],
                               kernel = kernel, rho = rho, n_threads = n_threads)
        R_tmp = nystrom_kernel(K_mm_tmp, K_nm_tmp, m_in = m, n_threads = n_threads)$R
        reml_lowrank(R_tmp, y, lambda, n_threads = n_threads)
      }else if(opt == "rff"){
        rand_tmp = rff_random(m = m, rho = rho, d = d, kernel = kernel)
        Z_tmp = make_Z(x, rand_tmp$W, rand_tmp$b, n_threads = n_threads)
        reml_lowrank(Z_tmp, y, lambda, n_threads = n_threads)
      }

    }else{
      if(selection_method == "fastCV")
        CVSTmethod = function(...) CVST::fastCV(...,
                                                setup = CVST::constructCVSTModel(),
                                                verbose = verbose)
      else
        CVSTmethod = function(...) CVST::CV(..., verbose = verbose)

      data = CVST::constructData(x, y)
      if(opt == "exact"){
        ojct = CVST::constructLearner(krr_fit_exact, krr_pred)
        param_sets = CVST::constructParams(kernel = kernel, rho = rho, lambda = lambda,
                                           n_threads = n_threads, verbose = verbose)
      }else if(opt == "pivoted"){
        ojct = CVST::constructLearner(krr_fit_pivoted, krr_pred)
        param_sets = CVST::constructParams(kernel = kernel, m = m, eps = eps,
                                           rho = rho, lambda = lambda,
                                           n_threads = n_threads, verbose = verbose)
      }else if(opt == "nystrom"){
        ojct = CVST::constructLearner(krr_fit_nystrom, krr_pred)
        param_sets = CVST::constructParams(kernel = kernel, m = m,
                                           rho = rho, lambda = lambda,
                                           n_threads = n_threads, verbose = verbose)
      }else if(opt == "rff"){
        ojct = CVST::constructLearner(krr_fit_rff, krr_pred_rff)
        param_sets = CVST::constructParams(kernel = kernel, m = m,
                                           rho = rho, lambda = lambda,
                                           n_threads = n_threads, verbose = verbose)
      }
      best_param = CVSTmethod(data, ojct, param_sets)
      lambda = best_param[[1]]$lambda
    }
  }


  # Fitting
  result_values = list()
  class(result_values) = "krr"
  result_values$removed_row_idx = removed_row_idx

  if(opt == "rff"){
    rand_set = rff_random(m = m, rho = rho, d = d, kernel = kernel)
    rslt = rff(x, y, rand_set$W, rand_set$b, lambda = lambda, n_threads = n_threads)

    result_values$coefficients = rslt$coefficients
    result_values$fitted.values = as.vector(rslt$fitted.values)
    result_values$opt = opt
    result_values$kernel = kernel
    result_values$x = x
    result_values$y = y
    result_values$lambda = lambda
    result_values$rho = rho
    result_values$n_threads = n_threads
    result_values$selection_method = selection_method
    result_values$call = call
    result_values$m = m
    result_values$approx_factor = rslt$Z
    result_values$W = rslt$W
    result_values$b = rslt$b
    return(result_values)

  }else if(opt == "exact"){
    K = make_kernel(x, kernel = kernel, rho = rho, n_threads = n_threads)

    rslt = solve_chol(K + diag(n * lambda, n), y)

    result_values$coefficients = rslt$coefficients
    result_values$fitted.values = as.vector(K %*% rslt$coefficients)
    result_values$opt = opt
    result_values$kernel = kernel
    result_values$x = x
    result_values$y = y
    result_values$lambda = lambda
    result_values$rho = rho
    result_values$n_threads = n_threads
    result_values$selection_method = selection_method
    result_values$call = call
    result_values$K = K
    result_values$chol_factor = rslt$chol_factor
    return(result_values)

  }else if(opt == "pivoted"){
    rslt = pchol(x, y, lambda = lambda, rho = rho,
                 kernel = kernel, m = m, verbose = verbose)

    result_values$coefficients = rslt$coefficients
    result_values$fitted.values = as.vector(rslt$fitted.values)
    result_values$opt = opt
    result_values$kernel = kernel
    result_values$x = x
    result_values$y = y
    result_values$lambda = lambda
    result_values$rho = rho
    result_values$n_threads = n_threads
    result_values$selection_method = selection_method
    result_values$call = call
    result_values$m = rslt$m
    result_values$approx_factor = rslt$PR
    result_values$eps = eps
    return(result_values)

  }else if(opt == "nystrom"){
    if(is.null(idx_ny)) idx_ny = sample(seq_len(n), m)

    K_nm = make_kernel(x[idx_ny, , drop = FALSE], x,
                       kernel = kernel, rho = rho, n_threads = n_threads)

    K_mm = make_kernel(x[idx_ny, , drop = FALSE],
                       kernel = kernel, rho = rho, n_threads = n_threads)

    rslt = nystrom(K_nm = K_nm, K_mm = K_mm, m = m, y = y,
                   lambda = lambda, n_threads = n_threads)

    result_values$coefficients = as.vector(rslt$coefficients)
    result_values$fitted.values = as.vector(rslt$fitted.values)
    result_values$opt = opt
    result_values$kernel = kernel
    result_values$x = x
    result_values$y = y
    result_values$lambda = lambda
    result_values$rho = rho
    result_values$n_threads = n_threads
    result_values$selection_method = selection_method
    result_values$call = call

    result_values$m = rslt$m
    result_values$approx_factor = rslt$R
    return(result_values)
  }
}

krr_fit_exact = function(data, param) {
  x = data$x
  y = data$y
  n = nrow(x)

  lambda = as.numeric(param$lambda)

  K = make_kernel(x, kernel = param$kernel, rho = param$rho, n_threads = param$n_threads)
  rslt <- solve_chol(K + diag(n * lambda, n), y)

  return(list(data = data, kernel = param$kernel,
              coefficients = rslt$coefficients,
              rho = param$rho, lambda = param$lambda,
              n_threads = param$n_threads))

}

krr_fit_pivoted = function(data, param) {
  x = data$x
  y = data$y
  n = nrow(x)
  m = as.integer(min(n, param$m))  # as.integer(n * param$rate)
  lambda = as.numeric(param$lambda)

  rslt = pchol(x, y, lambda = lambda, rho = param$rho,
               kernel = param$kernel, m = m,
               eps = param$eps, verbose = FALSE)
  coefficients = rslt$coefficients

  return(list(data = data, kernel = param$kernel,
              coefficients = coefficients,
              rho = param$rho, lambda = param$lambda,
              verbose = param$verbose,
              n_threads = param$n_threads))

}


krr_fit_nystrom = function(data, param) {
  x = data$x
  y = data$y
  n = nrow(x)
  d = ncol(x)

  m = as.integer(min(n, param$m))  # as.integer(n * param$rate)

  lambda = as.numeric(param$lambda)

  idx_ny = sample(n, m)

  K_nm = make_kernel(x[idx_ny, , drop = FALSE], x,
                     kernel = param$kernel, rho = param$rho, n_threads = param$n_threads)

  K_mm = make_kernel(x[idx_ny, , drop = FALSE], x[idx_ny, , drop = FALSE],
                     kernel = param$kernel, rho = param$rho, n_threads = param$n_threads)

  rslt = nystrom(K_nm = K_nm, K_mm = K_mm, m = m, y = y,
                 lambda = lambda, n_threads = param$n_threads)

  coefficients = rslt$coefficients

  return(list(data = data, kernel = param$kernel,
              coefficients = coefficients,
              rho = param$rho, lambda = param$lambda,
              n_threads = rslt$n_threads))

}

krr_fit_rff = function(data, param) {
  x = data$x
  y = data$y
  n = nrow(x)
  d = ncol(x)

  lambda = as.numeric(param$lambda)
  rand_set = rff_random(m = as.integer(min(n, param$m)), d = d, param$rho, kernel = param$kernel)

  rslt = rff(x, y, rand_set$W, rand_set$b, lambda = lambda, n_threads = param$n_threads)
  coefficients = rslt$coefficients

  return(list(data = data,
              kernel = param$kernel,
              coefficients = rslt$coefficients,
              Z = rslt$Z,
              m = rslt$m,
              W = rslt$W,
              b = rslt$b,
              rho = param$rho, lambda = param$lambda,
              n_threads = rslt$n_threads))

}


krr_pred = function(model, newdata) {
  x = model$data$x
  x_new = newdata$x
  rho = model$rho
  kernel = model$kernel
  opt = model$opt

  K_new = make_kernel(x, x_new, kernel = kernel, rho = rho, n_threads = model$n_threads)


  return(as.vector(K_new %*% model$coefficients))
}


krr_pred_rff = function(model, newdata) {
  x_new = newdata$x
  W = model$W
  b = model$b
  coef = model$coefficients
  Z_new = make_Z(x_new, W = W, b = b)

  return(as.vector(Z_new %*% coef))
}

test_that("fastkrr fits correctly across all computation options (exact, nystrom, pivoted, rff)", {
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  # 1. Exact option & S3 object structure verification
  fit_exact <- fastkrr(data = df, response = "y", kernel = "gaussian", opt = "exact", lambda = 1e-4)
  expect_s3_class(fit_exact, "krr")
  expect_type(fit_exact$coefficients, "double")
  expect_equal(length(fit_exact$coefficients), n)
  expect_equal(length(fit_exact$fitted.values), n)
  expect_false(is.null(fit_exact$chol_factor))

  # 2. Nystrom option
  fit_nys <- fastkrr(data = df, response = "y", kernel = "gaussian", opt = "nystrom", lambda = 1e-4, m = 10)
  expect_equal(length(fit_nys$fitted.values), n)
  expect_false(is.null(fit_nys$approx_factor))

  # 3. Pivoted Cholesky option
  fit_piv <- fastkrr(data = df, response = "y", kernel = "gaussian", opt = "pivoted", lambda = 1e-4, m = 10)
  expect_equal(length(fit_piv$fitted.values), n)
  expect_false(is.null(fit_piv$approx_factor))

  # 4. RFF option
  fit_rff <- fastkrr(data = df, response = "y", kernel = "gaussian", opt = "rff", lambda = 1e-4, m = 10)
  expect_equal(length(fit_rff$fitted.values), n)
  expect_equal(length(fit_rff$coefficients), 10) # m-dimensional weight vector
  expect_false(is.null(fit_rff$approx_factor))
})

test_that("predict.krr handles both missing newdata and out-of-sample predictions", {
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  fit <- fastkrr(data = df, response = "y", kernel = "gaussian", opt = "exact", lambda = 1e-4)

  # Returns fitted values when newdata is missing
  expect_equal(predict(fit), fit$fitted.values)

  # Evaluates predictions on new input design matrix
  new_X <- matrix(runif(10 * d), nrow = 10, ncol = d)
  colnames(new_X) <- c("x1", "x2")

  preds <- predict(fit, new_X)
  expect_equal(length(preds), 10)
  expect_false(anyNA(preds))

  # Evaluates predictions for RFF option
  fit_rff <- fastkrr(data = df, response = "y", kernel = "gaussian", opt = "rff", lambda = 1e-4, m = 8)
  preds_rff <- predict(fit_rff, new_X)
  expect_equal(length(preds_rff), 10)
  expect_false(anyNA(preds_rff))
})

test_that("lambda specification and selection_method behave correctly", {
  set.seed(123)
  n <- 25
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  # 1. REML tuning with length-2 min/max vector
  fit_reml <- fastkrr(data = df, response = "y", kernel = "gaussian", opt = "exact",
                      selection_method = "REML", lambda = c(1e-8, 1e-2))
  expect_type(fit_reml$lambda, "double")
  expect_length(fit_reml$lambda, 1)

  # 2. Error when passing length-2 vector to CV methods
  expect_error(
    fastkrr(data = df, response = "y", selection_method = "exactCV", lambda = c(1e-4, 1e-2)),
    "only supported when selection_method = 'REML'"
  )

  # 3. Error when passing length >= 3 grid to REML
  expect_error(
    fastkrr(data = df, response = "y", selection_method = "REML", lambda = c(1e-4, 1e-3, 1e-2)),
    "REML does not support a lambda grid of length >= 3"
  )
})

test_that("na.rm argument handles missing values appropriately", {
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)

  # Inject missing value
  X[5, 1] <- NA
  df <- data.frame(X, y = y)

  # Throws error when na.rm = FALSE
  expect_error(
    fastkrr(data = df, response = "y", kernel = "gaussian", opt = "exact", lambda = 1e-4, na.rm = FALSE),
    "Missing values"
  )

  # Removes missing rows gracefully when na.rm = TRUE
  expect_no_error({
    fit_na <- fastkrr(data = df, response = "y", kernel = "gaussian", opt = "exact", lambda = 1e-4, na.rm = TRUE)
  })
  expect_equal(fit_na$removed_row_idx, 5)
  expect_equal(length(fit_na$fitted.values), n - 1)
})

test_that("coef.krr extracts model coefficients correctly as a numeric vector", {
  # Set seed for reproducibility
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  # Fit a FastKRR model
  model <- fastkrr(data = df, response = "y", kernel = "gaussian",
                   opt = "exact", rho = 1.0, lambda = 1e-3)

  # Extract coefficients using coef()
  extracted_coefs <- coef(model)

  # Validate output type, dimension, and structure
  expect_type(extracted_coefs, "double")
  expect_true(is.vector(extracted_coefs))
  expect_equal(length(extracted_coefs), length(model$coefficients))
  expect_equal(extracted_coefs, as.vector(model$coefficients))
})

test_that("coef.krr works across different approximation options", {
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  # Test for RFF option where coefficients represent beta weights
  model_rff <- fastkrr(data = df, response = "y", kernel = "gaussian",
                       opt = "rff", m = 10, rho = 1.0, lambda = 1e-3)

  extracted_coefs_rff <- coef(model_rff)

  expect_type(extracted_coefs_rff, "double")
  expect_true(is.vector(extracted_coefs_rff))
  expect_equal(length(extracted_coefs_rff), 10)
})

test_that("param.default throws an informative error for unsupported classes", {
  # Expect error when passing unsupported object type (e.g., numeric vector or lm)
  expect_error(
    param("invalid_object"),
    "No 'param' method for objects of class: character"
  )

  dummy_list <- list(a = 1)
  class(dummy_list) <- "unsupported_class"
  expect_error(
    param(dummy_list),
    "No 'param' method for objects of class: unsupported_class"
  )
})

test_that("param.krr extracts hyperparameters and returns a krr_params object invisibly", {
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  # Fit model using fastkrr
  fit <- fastkrr(
    data = df,
    response = "y",
    kernel = "gaussian",
    opt = "pivoted",
    rho = 1.0,
    lambda = 1e-4,
    n_threads = 1
  )

  # Check invisible return value structure and class
  res <- param(fit)
  expect_s3_class(res, "krr_params")
  expect_type(res, "list")
  expect_equal(res$kernel, "gaussian")
  expect_equal(res$opt, "pivoted")
  expect_equal(res$rho, 1.0)
  expect_equal(res$lambda, 1e-4)
})

test_that("param.krr prints human-readable output to the console", {
  set.seed(123)
  n <- 20
  d <- 1
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  y <- X[, 1] * 2 + rnorm(n, sd = 0.05)
  df <- data.frame(X, y = y)

  fit <- fastkrr(
    data = df,
    response = "y",
    kernel = "gaussian",
    opt = "exact",
    lambda = 1e-3
  )

  # Verify console print output contains Call and Parameters sections
  expect_output(param(fit), "Call:")
  expect_output(param(fit), "Parameters:")
})

test_that("print.krr displays model details and invisibly returns the original object", {
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X) <- c("x1", "x2")
  y <- sin(X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  # Fit a FastKRR model
  fit <- fastkrr(
    data = df,
    response = "y",
    kernel = "gaussian",
    opt = "exact",
    rho = 1.0,
    lambda = 1e-3
  )

  # 1. Check console output text patterns
  expect_output(print(fit), "Call:")
  expect_output(print(fit), "Parameters:")

  # 2. Verify that print.krr invisibly returns the input krr object
  res <- print(fit)
  expect_s3_class(res, "krr")
  expect_equal(res, fit)
})

test_that("print.approx_kernel displays approximation details to console", {
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)

  # Construct an approximated kernel object
  K_approx <- approx_kernel(X = X, opt = "nystrom", m = 10, rho = 1.0, n_threads = 1)

  # Verify console output text patterns
  expect_output(print(K_approx), "Call:")
  expect_output(print(K_approx), "Options:")

  # Capture output dataframe returned by print.approx_kernel
  res_df <- print(K_approx)
  expect_s3_class(res_df, "data.frame")
  expect_true("opt" %in% names(res_df))
})

test_that("plot.krr returns a valid ggplot object for 1D input", {
  # Generate 1D synthetic dataset
  set.seed(123)
  n <- 30
  X <- matrix(runif(n), nrow = n, ncol = 1)
  colnames(X) <- "x1"
  y <- sin(2 * pi * X[, 1]) + rnorm(n, sd = 0.1)
  df <- data.frame(X, y = y)

  # Fit model
  fit <- fastkrr(data = df, response = "y", kernel = "gaussian",
                 opt = "exact", lambda = 1e-3)

  # Test plot with points (default)
  p1 <- plot(fit, show_points = TRUE)
  expect_s3_class(p1, "ggplot")
  expect_equal(length(p1$layers), 2) # geom_point + geom_line

  # Test plot without points
  p2 <- plot(fit, show_points = FALSE)
  expect_s3_class(p2, "ggplot")
  expect_equal(length(p2$layers), 1) # geom_line only
})

test_that("plot.krr throws informative errors on invalid inputs or d >= 2", {
  set.seed(123)

  # Case 1: Input is not a 'krr' object
  expect_error(
    plot.krr("not_a_krr_object"),
    "Input must be a fitted model of class 'krr'."
  )

  # Case 2: Multi-dimensional input matrix (d >= 2)
  n <- 30
  d <- 2
  X_2d <- matrix(runif(n * d), nrow = n, ncol = d)
  colnames(X_2d) <- c("x1", "x2")
  y <- X_2d[, 1] + X_2d[, 2] + rnorm(n, sd = 0.1)
  df_2d <- data.frame(X_2d, y = y)

  fit_2d <- fastkrr(data = df_2d, response = "y", kernel = "gaussian",
                    opt = "exact", lambda = 1e-3)

  # Expect error for multivariate inputs
  expect_error(
    plot(fit_2d),
    "plot.krr\\(\\) currently supports only 1D input."
  )
})

#' Kernel Ridge Regression
#'
#' @description
#' Defines a Kernel Ridge Regression model specification
#' for use with the tidymodels ecosystem via \pkg{parsnip}. This spec can be
#' paired with the \code{"fastkrr"} engine implemented in this package to fit
#' exact or kernel approximation (Nyström, Pivoted Cholesky, Random Fourier Features) within
#' \pkg{recipes}/\pkg{workflows} pipelines.
#'
#'
#' @param mode A single string; only `"regression"` is supported.
#' @param kernel Kernel matrix has two kinds of Kernel ("gaussian", "laplace").
#' @param opt Method for constructing or approximating :
#'  \describe{
#'   \item{\code{"exact"}}{Construct the full kernel matrix
#'   \eqn{K \in \mathbb{R}^{n\times n}} using design matrix \eqn{X}.}
#'   \item{\code{"nystrom"}}{Construct a low-rank approximation of
#'       the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
#'       using the Nyström approximation.}
#'   \item{\code{"pivoted"}}{Construct a low-rank approximation of
#'       the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
#'       using Pivoted Cholesky decomposition.}
#' \item{\code{"rff"}}{Use Random Fourier Features to construct a feature map
#'   \eqn{Z \in \mathbb{R}^{n \times m}} (with \eqn{m} random features) so that
#'   \eqn{K \approx Z Z^\top}. Here, \eqn{m} is the number of features.}

#'  }
#' @param m  Approximation rank(number of random features) used for the low-rank kernel approximation.
#' @param eps Tolerance parameter used only in \code{"pivoted"}
#'   for stopping criterion of the Pivoted Cholesky decomposition.
#' @param n_threads Number of parallel threads. It is applied only for
#'  \code{opt = "nystrom"} or \code{opt = "rff"}, and for the
#'   Laplace kernel (\code{kernel = "laplace"}).
#' @param rho  Scaling parameter of the kernel(\eqn{\rho}).
#' @param penalty Regularization parameter. It must be a positive numeric
#' value, a numeric vector containing at least three positive values, or
#' \code{tune()}. When using the \pkg{tidymodels} interface for
#' hyperparameter selection, setting \code{penalty = tune()} is recommended.
#' Candidate values are then supplied through the \code{grid} argument of
#' \code{tune_grid()} and evaluated using the resampling scheme specified
#' by the user.
#' @param na.rm Logical. If \code{TRUE}, rows containing missing values are
#'   removed before fitting. Defaults to \code{FALSE}.
#'
#'
#' @return A parsnip model specification of class \code{"krr_reg"}.
#'
#' @examples
#' \donttest{
#' if (all(vapply(
#'   c("parsnip","stats","modeldata"),
#'   requireNamespace, quietly = TRUE, FUN.VALUE = logical(1)
#' ))) {
#' library(tidymodels)
#' library(parsnip)
#' library(stats)
#' library(modeldata)
#'
#' # Data analysis
#' data(ames)
#' ames = ames %>% mutate(Sale_Price = log10(Sale_Price))
#'
#' set.seed(502)
#' ames_split = initial_split(ames, prop = 0.80, strata = Sale_Price)
#' ames_train = training(ames_split) # dim (2342, 74)
#' ames_test  = testing(ames_split) # dim (588, 74)
#'
#' # Model spec
#' krr_spec = krr_reg(kernel = "gaussian", opt = "nystrom",
#'                    m = 50, eps = 1e-6, n_threads = 1,
#'                    rho = 1, penalty = tune()) %>%
#'  set_engine("fastkrr") %>%
#'  set_mode("regression")
#'
#' # Define rec
#' rec = recipe(Sale_Price ~ Longitude + Latitude, data = ames_train)
#'
#' # workflow
#' wf = workflow() %>%
#'   add_recipe(rec) %>%
#'   add_model(krr_spec)
#'
#' # Define hyper-parameter grid
#' param_grid = grid_regular(
#'   dials::penalty(range = c(-10, -3)),
#'   levels = 5
#' )
#'
#' # CV setting
#' set.seed(123)
#' cv_folds = vfold_cv(ames_train, v = 5, strata = Sale_Price)
#'
#' # Tuning
#' tune_results = tune_grid(
#'   wf,
#'   resamples = cv_folds,
#'   grid = param_grid,
#'   metrics = metric_set(rmse),
#'   control = control_grid(verbose = TRUE, save_pred = TRUE)
#' )
#'
#' # Result check
#' collect_metrics(tune_results)
#'
#' # Select best parameter
#' best_params = select_best(tune_results, metric = "rmse")
#'
#' # Finalized model spec using best parameter
#' final_spec = finalize_model(krr_spec, best_params)
#' final_wf = workflow() %>%
#'   add_recipe(rec) %>%
#'   add_model(final_spec)
#'
#' # Finalized fitting using best parameter
#' final_fit = final_wf %>% fit(data = ames_train)
#'
#' # Prediction
#' predict(final_fit, new_data = ames_test)
#' print(best_params)
#'
#' }}
#'
#' @export
krr_reg = function(mode = "regression", kernel = NULL, opt = NULL, eps = NULL,
                   n_threads = NULL, m = NULL, rho = NULL, penalty = NULL,
                   na.rm = NULL) {
  if (mode != "regression") {
    rlang::abort("`mode` should be 'regression'.")
  }
  args = list(
    kernel  = rlang::enquo(kernel),
    opt     = rlang::enquo(opt),
    m       = rlang::enquo(m),
    eps      = rlang::enquo(eps),
    n_threads = rlang::enquo(n_threads),
    rho     = rlang::enquo(rho),
    penalty = rlang::enquo(penalty),
    na.rm = rlang::enquo(na.rm)
  )
  parsnip::new_model_spec(
    "krr_reg",
    args     = args,
    eng_args = NULL,
    mode     = mode,
    method   = NULL,
    engine   = NULL
  )
}

#' Expose tunable parameters for \code{"krr_reg"}
#'
#' @description
#' Supplies a tibble of tunable arguments for \code{"krr_reg()"}.
#'
#' @param x A \code{"krr_reg"} model specification.
#' @param ... Not used; included for S3 method compatibility.
#'
#' @return A tibble (one row per tunable parameter)
#'    with columns \code{"name"}, \code{"call_info"}, \code{"source"},
#'   \code{"component"}, and \code{"component_id"}.
#'
#' @importFrom generics tunable
#' @importFrom tibble tibble
#' @export
tunable.krr_reg = function(x, ...) {
  tibble::tibble(
    name = "penalty",
    call_info = list(list(pkg = "dials", fun = "penalty", range = c(-10, -3))),
    source = "model_spec",
    component = "krr_reg",
    component_id = "main"
  )
}


# ===================================================
# helper: remove an existing parsnip model registration
# ===================================================
.reset_parsnip_model = function(model) {
  if (!requireNamespace("parsnip", quietly = TRUE)) return(invisible())
  get_model_env = get("get_model_env", envir = asNamespace("parsnip"))
  env = get_model_env()
  if (exists(model, envir = env, inherits = FALSE)) {
    rm(list = model, envir = env)
  }
}

# =========================
# parsnip registration
# =========================
make_krr_reg = function(){
  # Idempotent guard: do nothing if already registered
  get_model_env = get("get_model_env", envir = asNamespace("parsnip"))
  env = get_model_env()
  if (exists("krr_reg", envir = env, inherits = FALSE)) {
    return(invisible(NULL))
  }

  # 1) model/mode
  parsnip::set_new_model("krr_reg")
  parsnip::set_model_mode(model = "krr_reg", mode = "regression")

  # 2) engine
  parsnip::set_model_engine(model = "krr_reg", mode = "regression", eng = "fastkrr")
  parsnip::set_dependency(model = "krr_reg", eng = "fastkrr", pkg = "FastKRR")

  # 3) parameter mapping
  parsnip::set_model_arg(
    model    = "krr_reg",
    eng      = "fastkrr",
    parsnip  = "penalty",
    original = "lambda",
    func     = list(pkg = "dials", fun = "penalty"),# func = list(pkg = "base", fun = "identity"),
    has_submodel = FALSE
  )

  for (nm in c("kernel", "opt", "m", "eps", "n_threads", "rho", "na.rm")){
    parsnip::set_model_arg(
      model    = "krr_reg",
      eng      = "fastkrr",
      parsnip  = nm,
      original = nm,
      func     = list(pkg = "base", fun = "identity"),
      has_submodel = FALSE
    )
  }

  # 4) fit function mapping
  #    fastkrr(x, y, kernel, opt, m, rho, lambda, selection_method)
  parsnip::set_fit(
    model = "krr_reg",
    eng   = "fastkrr",
    mode  = "regression",
    value = list(
      interface = "matrix",
      protect   = c("x", "y"),
      func      = c(pkg = "FastKRR", fun = "fastkrr_fit_wrapper"),
      args      = list(
        x      = rlang::expr(x),
        y      = rlang::expr(y),
        kernel = rlang::expr(kernel),
        opt    = rlang::expr(opt),
        m   = rlang::expr(m),
        eps  = rlang::expr(eps),
        n_threads  = rlang::expr(n_threads),
        rho    = rlang::expr(rho),
        lambda = rlang::expr(penalty),
        na.rm = rlang::expr(na.rm)
      ),
      defaults = list()
    )
  )

  # 5) prediction function mapping
  #    predict(model, newdata) -> numeric vector
  parsnip::set_pred(
    model = "krr_reg",
    eng   = "fastkrr",
    mode  = "regression",
    type  = "numeric",
    value = list(
      pre  = NULL,
      post = NULL,
      func = c(fun = "predict"),
      args = list(
        object = rlang::expr(object$fit),  ### model   = rlang::expr(object$fit)
        newdata = rlang::expr(new_data)
      )
    )
  )

  # 6) encoding
  parsnip::set_encoding(
    model = "krr_reg",
    eng   = "fastkrr",
    mode  = "regression",
    options = list(
      predictor_indicators = "none",
      compute_intercept    = FALSE,
      remove_intercept     = FALSE,
      allow_sparse_x       = FALSE
    )
  )
}

# =========================
# package load hook — force reset and re-register on every load
# =========================
.onLoad = function(libname, pkgname) {
  if (!requireNamespace("parsnip", quietly = TRUE)) return(invisible())


  # .reset_parsnip_model("krr_reg")

  make_krr_reg()

}



# =========================
# optional: spec updater
# =========================
#' @importFrom stats update
#' @exportS3Method update krr_reg
update.krr_reg = function(object, parameters = NULL,
                          kernel = NULL, opt = NULL,
                          m = NULL, eps = NULL,
                          n_threads = NULL, rho = NULL,
                          penalty = NULL, na.rm = NULL,
                          fresh = FALSE, ...) {

  # if (requireNamespace("parsnip", quietly = TRUE) &&
  #     "update.model_spec" %in% getNamespaceExports("parsnip")) {
  #   return(stats::update(
  #     object,
  #     kernel = kernel, opt = opt, m = m, eps = eps, n_threads = n_threads,
  #     rho = rho, selection_method = selection_method, penalty = penalty,
  #     parameters = parameters, fresh = fresh, ...
  #   ))
  # }

  args_new = list()

  if (!missing(kernel)) {
    args_new$kernel = rlang::enquo(kernel)
  }

  if (!missing(opt)) {
    args_new$opt = rlang::enquo(opt)
  }

  if (!missing(m)) {
    args_new$m = rlang::enquo(m)
  }

  if (!missing(eps)) {
    args_new$eps = rlang::enquo(eps)
  }

  if (!missing(n_threads)) {
    args_new$n_threads = rlang::enquo(n_threads)
  }

  if (!missing(rho)) {
    args_new$rho = rlang::enquo(rho)
  }

  if (!missing(na.rm)) {
    args_new$na.rm = rlang::enquo(na.rm)
  }

  if (!missing(penalty)) {
    args_new$penalty = rlang::enquo(penalty)
  }

  args = if (isTRUE(fresh)) {
    args_new
  } else {
    object$args[names(args_new)] = args_new
    object$args
  }
  object$args = if (fresh) args_new else utils::modifyList(object$args, args_new)
  object
}





#' Internal bridge for parsnip fit interface
#' @keywords internal
#' @export
fastkrr_fit_wrapper = function(x, y,
                               kernel = "gaussian",
                               opt = "exact",
                               m = NULL,
                               eps = 1e-6,
                               rho = 1,
                               lambda = NULL,
                               n_threads = 4,
                               verbose = FALSE,
                               na.rm = FALSE) {

  df = as.data.frame(x)

  response_col = ".outcome"
  while (response_col %in% names(df)) {
    response_col = paste0(response_col, "_")
  }
  df[[response_col]] = y

  fastkrr(
    data              = df,
    response          = response_col,
    kernel            = kernel,
    opt               = opt,
    m                 = m,
    eps               = eps,
    rho               = rho,
    lambda            = lambda,
    n_threads         = n_threads,
    verbose           = verbose,
    na.rm             = na.rm
  )
}












utils::globalVariables(c(
  "x","y","opt","m", "eps","n_threads","rho","penalty", "na.rm", "object","new_data"
))

test_that("krr_reg initializes correctly and validates mode", {
  skip_if_not_installed("parsnip")

  # 1. Valid specification creation
  spec <- krr_reg(
    kernel = "gaussian",
    opt = "exact",
    penalty = 0.01,
    rho = 1.0
  )
  expect_s3_class(spec, "krr_reg")
  expect_s3_class(spec, "model_spec")
  expect_equal(spec$mode, "regression")

  # 2. Invalid mode error handling check
  expect_error(
    krr_reg(mode = "classification"),
    "`mode` should be 'regression'"
  )
})

test_that("tunable.krr_reg returns appropriate parameter information", {
  skip_if_not_installed("parsnip")
  skip_if_not_installed("generics")

  spec <- krr_reg()
  tb <- generics::tunable(spec)

  expect_s3_class(tb, "tbl_df")
  expect_equal(tb$name, "penalty")
  expect_equal(tb$component, "krr_reg")
})

test_that("update.krr_reg modifies parameters correctly with fresh = FALSE/TRUE", {
  skip_if_not_installed("parsnip")

  spec <- krr_reg(kernel = "gaussian", opt = "exact", rho = 1.0)

  # 1. Update individual parameters (fresh = FALSE)
  spec_updated <- update(
    spec,
    kernel = "laplace",
    opt = "nystrom",
    m = 20,
    eps = 1e-4,
    n_threads = 2,
    rho = 2.0,
    penalty = 0.05,
    na.rm = TRUE,
    fresh = FALSE
  )

  expect_s3_class(spec_updated, "krr_reg")
  expect_equal(rlang::quo_get_expr(spec_updated$args$kernel), "laplace")
  expect_equal(rlang::quo_get_expr(spec_updated$args$opt), "nystrom")

  # 2. Update with fresh = TRUE
  spec_fresh <- update(spec_updated, kernel = "gaussian", fresh = TRUE)
  expect_true("kernel" %in% names(spec_fresh$args))
  expect_false("opt" %in% names(spec_fresh$args)) # previous args cleared
})

test_that("make_krr_reg and helper functions handle registration idempotently", {
  skip_if_not_installed("parsnip")

  # Ensure model is registered
  make_krr_reg()

  # Calling again should do nothing and return NULL safely
  res <- make_krr_reg()
  expect_null(res)

  # Check helper function without breaking parsnip environment
  expect_invisible(.reset_parsnip_model("non_existent_model_xyz"))
})

test_that("fastkrr_fit_wrapper handles dataframe conversions and outcome name collisions", {
  set.seed(123)
  n <- 25
  X <- matrix(runif(n * 2), nrow = n, ncol = 2)
  y <- rnorm(n)

  # 1. Normal matrix fitting via wrapper
  fit_res <- fastkrr_fit_wrapper(
    x = X,
    y = y,
    kernel = "gaussian",
    opt = "exact",
    rho = 1.0,
    lambda = 0.01
  )
  expect_s3_class(fit_res, "krr")

  # 2. Name collision test (dataframe already containing `.outcome` column)
  df_x <- data.frame(x1 = runif(n), .outcome = runif(n))
  fit_col_test <- fastkrr_fit_wrapper(
    x = df_x,
    y = y,
    kernel = "gaussian",
    opt = "exact",
    rho = 1.0,
    lambda = 0.01
  )
  expect_s3_class(fit_col_test, "krr")
})

test_that("parsnip workflow integration (fit and predict) functions seamlessly", {
  skip_if_not_installed("parsnip")

  # Ensure model specification is cleanly registered before fitting
  make_krr_reg()

  set.seed(123)
  df_train <- data.frame(
    x1 = runif(30),
    x2 = runif(30),
    y  = rnorm(30)
  )

  df_test <- data.frame(
    x1 = runif(5),
    x2 = runif(5)
  )

  # Set parsnip model specification
  krr_spec <- krr_reg(
    kernel = "gaussian",
    opt = "exact",
    penalty = 0.01,
    rho = 1.0
  ) %>%
    parsnip::set_engine("fastkrr") %>%
    parsnip::set_mode("regression")

  # Fit parsnip model
  fit_parsnip <- parsnip::fit(krr_spec, y ~ x1 + x2, data = df_train)
  expect_s3_class(fit_parsnip, "model_fit")

  # Predict via parsnip
  preds <- predict(fit_parsnip, new_data = df_test)
  expect_s3_class(preds, "tbl_df")
  expect_equal(nrow(preds), 5)
  expect_true(".pred" %in% names(preds))
})

#' Print method for fitted Kernel Ridge Regression models
#'
#' @description
#' Displays a concise summary of key information from a fitted Kernel Ridge Regression (KRR) model,
#' including the original function call and the main hyperparameter options used during fitting.
#'
#' @param x An S3 object of class \code{krr}, typically returned by
#'   \code{\link{fastkrr}}.
#' @param ... Additional arguments (currently ignored).
#'
#' @return Invisibly returns the input \code{krr} object after printing the summary to the console.
#'
#' @seealso \code{\link{fastkrr}}, \code{\link{print.approx_kernel}}
#'
#' @examples
#' # Data setting
#' set.seed(1)
#' lambda = 1e-4
#' d = 1
#' n = 50
#' rho = 1
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
#' y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)
#'
#' data = data.frame(X, y = y)
#'
#' # Example: exact
#' model = fastkrr(data = data, response = "y",
#'                  kernel = "gaussian", opt = "exact",
#'                  rho = rho, lambda = lambda)
#'
#' print(model)
#' @export
print.krr = function(x, ...) {
  model = x

  cat("Call:\n")
  print(model$call)
  cat("\n")

  keys = c("kernel", "opt", "selection_method",
           "rho", "lambda", "m", "eps", "n_threads")
  present = intersect(keys, names(model))

  if (length(present) > 0) {
    df = as.data.frame(model[present], optional = TRUE)

    cat("Parameters:\n")
    print(df, row.names = FALSE)
    cat("\n")
  }

  invisible(model)
}


#' Print method for approximated kernel matrices
#'
#' @description
#' Displays the key algorithmic choices and hyperparameter options used to construct
#' an approximated kernel matrix object.
#'
#' @details
#' The function summarizes critical metadata attributes stored within the
#' \code{approx_kernel} object, including the approximation strategy (\code{opt}),
#' kernel type, approximation degree (\code{m}), numerical tolerance (\code{eps}),
#' and the number of parallel computational threads used.
#'
#' @param x An S3 object created by \code{\link{approx_kernel}}.
#' @param ... Additional arguments (currently ignored).
#'
#' @return Invisibly returns the input \code{approx_kernel} object after printing the options.
#'
#' @seealso \code{\link{approx_kernel}}, \code{\link{print.krr}}
#'
#' @examples
#' # Data setting
#' set.seed(1)
#' d = 1
#' n = 100
#' rho = 1
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
#'
#' # Example: nystrom
#' K_nystrom = approx_kernel(X = X, opt = "nystrom", rho = rho, n_threads = 1)
#'
#' print(K_nystrom)
#' @export
print.approx_kernel = function(x, ...){
  approx_object = x
  cat("Call:\n")
  print(approx_object$call)

  cat("\n")
  idx = names(approx_object) %in% c("opt", "kernel", "d", "rho", "m", "eps", "n_threads")
  df = as.data.frame(approx_object[idx])
  row.names(df) = ""
  cat("Options:\n")
  print(df)
}

#' Extract/print hyperparameters of fitted models
#'
#' @description
#' \code{param} is a generic S3 function that displays (and invisibly returns)
#' model hyperparameters. Methods are provided for \code{"krr"} objects.
#'
#' @param x An object.
#' @param ... Additional arguments passed to methods.
#' @return A named list of hyperparameters (invisibly); may print side effects.
#' @keywords internal
#' @export
param = function(x, ...) UseMethod("param")

#' @rdname param
#' @export
param.default = function(x, ...) {
  stop(sprintf("No 'param' method for objects of class: %s", paste(class(x), collapse=", ")))
}

#' Displays hyperparameters of fitted Kernel Ridge Regression models
#'
#' @description
#' Displays (and invisibly returns) the hyperparameters actually used by a
#' fitted object. For \code{krr} objects returned by \code{\link{fastkrr}},
#' this prints a concise hyperparameter panel (e.g., \code{rho}, \code{lambda},
#' \code{m}, \code{eps}, \code{n_threads}, \code{d}).
#'
#' @details
#' \strong{Pivoted approximation note}: When \code{opt = "pivoted"}, the
#' effective number of pivots \code{m} used during the approximation may be
#' smaller than the user-specified \code{m} because the algorithm can stop
#' early based on \code{eps}. If you want to confirm the initial \code{m}
#' that you set, please see the printed \emph{Call} (the original function
#' call shows your input arguments).
#'
#' @param x An object of class \code{"krr"}, typically returned by
#'   \code{\link{fastkrr}}.
#' @param ... Additional arguments.
#'
#' @return
#' Prints a human-readable panel to the console and invisibly returns a
#' named list of class \code{"krr_params"} containing the extracted hyperparameters:
#' \itemize{
#'   \item \code{kernel}: Kernel type used ("gaussian" or "laplace").
#'   \item \code{opt}: Kernel approximation method.
#'   \item \code{selection_method}: Lambda tuning method.
#'   \item \code{rho}: Kernel scaling parameter.
#'   \item \code{lambda}: Regularization parameter.
#'   \item \code{m}: Rank or number of random features used for approximation.
#'   \item \code{eps}: Tolerance parameter (for pivoted Cholesky).
#'   \item \code{n_threads}: Number of parallel threads used.
#' }
#'
#' @seealso \code{\link{fastkrr}}
#'
#' @examples
#' # Data setting
#' set.seed(1)
#' lambda = 1e-4
#' d = 1
#' n = 50
#' rho = 1
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
#' y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)
#'
#' data = data.frame(X, y = y)
#'
#' model = fastkrr(data = data, response = "y",
#'                  kernel="gaussian", opt="pivoted",
#'                  rho=1, lambda=lambda, n_threads = 1)
#'
#' # Inspect hyperparameters
#' param(model)
#'
#' @export
param.krr = function(x, ...) {
  model = x

  keys = c("kernel", "opt", "selection_method",
           "rho", "lambda", "m", "eps", "n_threads")
  present = intersect(keys, names(model))

  out = model[present]
  class(out) = c("krr_params", "list")

  # Call
  cat("Call:\n")
  print(model$call)
  cat("\n")

  # Parameters
  if (length(present) > 0) {
    df = as.data.frame(model[present], optional = TRUE)
    row.names(df) = ""
    cat("Parameters:\n")
    print(df)
    cat("\n")
  }

  invisible(out)
}

# Generate random Fourier features parameters
rff_random = function(m, d = d, rho = 1, kernel = "gaussian"){
  W = matrix(0, nrow = m, ncol = d)
  b = rep(0, m)

  if(kernel == 'gaussian'){
    W = matrix(rnorm(m * d, 0, sqrt(2 * rho)), nrow = m)
  }else if(kernel == "laplace"){
    W = matrix(rcauchy(m * d, location = 0, scale = 1/sqrt(2 * rho)), nrow = m)
  }

  b = runif(m, 0, 2 * pi)

  return(list("W" = W,
              "b" = b))
}


rff_kernel = function(X,
                      W = NULL,
                      b = NULL,
                      m = NULL,
                      rho = 1,
                      kernel = "gaussian")
{
  n = nrow(X)
  d = ncol(X)

  # m default: n/10 (min: 1)
  if (is.null(m)) m = n^(1/2) * log(d + 5)
  m = as.integer(max(1, floor(m)))

  if (is.null(W) || is.null(b)) {
    # Temporarily set global variable `d` because rff.random references it
    old_d_exists = exists("d", envir = .GlobalEnv, inherits = FALSE)
    if (old_d_exists) old_d = get("d", envir = .GlobalEnv)
    on.exit({
    }, add = TRUE)

    rb = rff_random(m = m, d = d, rho = rho, kernel = kernel)
    W = rb$W
    b = rb$b
  } else {

    if (ncol(W) != d) stop("ncol(W) must equal ncol(X).")
    if (length(b) != nrow(W)) stop("length(b) must equal nrow(W).")
    m = nrow(W)
  }

  Z = make_Z(X, W, b)

  K_approx = tcrossprod(Z)

  return(list(
    K_approx = K_approx,
    m = m,
    W = W,
    b = b
  ))
}

#'  Kernel matrix construction for given datasets
#'
#' Constructs a kernel matrix \eqn{K \in \mathbb{R}^{n' \times n}} given two
#' datasets \eqn{X \in \mathbb{R}^{n \times d}} and \eqn{X' \in \mathbb{R}^{n' \times d}},
#' where \eqn{x_i \in \mathbb{R}^d} and \eqn{x'_j \in \mathbb{R}^d} denote the
#' i-th and j-th rows of \eqn{X} and \eqn{X'}, respectively, and
#' \eqn{K_{ji}=\mathcal{K}(x_i, x'_j)} for a user-specified kernel.
#' Implemented in C++ via RcppArmadillo.
#'
#' @param X Design matrix \eqn{X \in \mathbb{R}^{n \times d}} (rows  \eqn{x_i \in \mathbb{R}^d}).
#' @param X_new  Second matrix \eqn{X' \in \mathbb{R}^{n' \times d}} (rows  \eqn{x'_j \in \mathbb{R}^d}).
#' If omitted, \eqn{X' = X} and \eqn{n' = n}.
#' @param kernel Kernel type; one of \code{"gaussian"} or \code{"laplace"}.
#' @param rho Kernel width parameter (\eqn{\rho > 0}). Default is 1.
#' @param n_threads Number of parallel threads. If \code{NULL}, it defaults to
#'   half of the available system processors (\code{max_threads \%/\% 2}).
#'   For these parallelizable cases, if the system has 3 or fewer processors, it automatically
#'   falls back to \code{1} thread; otherwise, it is capped at \code{max_threads - 1}.
#'   For all other settings, it is restricted to \code{1} thread.
#'   Parallelization (implemented in C++) is one of the main advantages
#'   of this package and is applied only for \code{"laplace"} kernels.
#'
#' @details
#' Gaussian kernel: \deqn{K_{ji}=\mathcal{K}(x_i,x_j)=\exp\!\big(-\rho\|x_i-x_j\|_2^2\big)}
#' Laplace kernel:  \deqn{K_{ji}=\mathcal{K}(x_i,x_j)=\exp\!\big(-\rho\|x_i-x_j\|_1\big)}
#'
#' @return
#' The computed kernel matrix. If \code{X_new} is \code{NULL}, the result is a symmetric matrix
#' \eqn{K_{ij} = \mathcal{K}(x_i, x_j)}, with \eqn{K \in \mathbb{R}^{n \times n}}.
#' Otherwise, the result is a rectangular matrix
#' \eqn{K'_{ji} = \mathcal{K}(x_i, x'_j)}, with \eqn{K' \in \mathbb{R}^{n' \times n}}.
#'
#' @examples
#' # Data setting
#' set.seed(1)
#' d = 1
#' rho = 1
#' n = 100
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
#'
#' # New design matrix
#' new_n = 150
#' new_X = matrix(runif(new_n*d, 0, 1), nrow = new_n, ncol = d)
#'
#' # Make kernel : Gaussian kernel
#' K = make_kernel(X, kernel = "gaussian", rho = rho) ## symmetric matrix
#' new_K = make_kernel(X, new_X, kernel = "gaussian", rho = rho) ## rectangular matrix
#'
#' # Make kernel : Laplace kernel
#' K = make_kernel(X, kernel = "laplace", rho = rho, n_threads = 1) ## symmetric matrix
#' new_K = make_kernel(X, new_X, kernel = "laplace", rho = rho, n_threads = 1) ## rectangular matrix
#'
#' @name make_kernel
#' @usage make_kernel(X,
#'             X_new = NULL,
#'             kernel = "gaussian",
#'             rho = 1,
#'             n_threads = NULL)
#' @export make_kernel
#' @useDynLib FastKRR, .registration = TRUE
#' @importFrom Rcpp evalCpp
NULL

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

make_kernel <- function(X, X_new = NULL, kernel = "gaussian", rho = 1.0, n_threads = NULL) {
    .Call(`_FastKRR_make_kernel`, X, X_new, kernel, rho, n_threads)
}

nystrom_kernel <- function(K_mm, K_nm, m_in = NULL, n_threads = 4L) {
    .Call(`_FastKRR_nystrom_kernel`, K_mm, K_nm, m_in, n_threads)
}

nystrom <- function(K_mm, K_nm, y, m, lambda, n_threads = 4L) {
    .Call(`_FastKRR_nystrom`, K_mm, K_nm, y, m, lambda, n_threads)
}

pchol_kernel <- function(X, rho = 1.0, kernel = "gaussian", m = NULL, eps = 1e-6, verbose = TRUE) {
    .Call(`_FastKRR_pchol_kernel`, X, rho, kernel, m, eps, verbose)
}

pchol <- function(X, y, lambda, rho = 1.0, kernel = "gaussian", m = NULL, eps = 1e-6, verbose = TRUE) {
    .Call(`_FastKRR_pchol`, X, y, lambda, rho, kernel, m, eps, verbose)
}

reml_exact <- function(K, y, lambda_vec, max_iter = 50L, tol = 1e-8, n_threads = 1L) {
    .Call(`_FastKRR_reml_exact`, K, y, lambda_vec, max_iter, tol, n_threads)
}

reml_lowrank <- function(F, y, lambda_vec, max_iter = 50L, tol = 1e-8, n_threads = 1L) {
    .Call(`_FastKRR_reml_lowrank`, F, y, lambda_vec, max_iter, tol, n_threads)
}

make_Z <- function(X, W, b, n_threads = 4L) {
    .Call(`_FastKRR_make_Z`, X, W, b, n_threads)
}

rff <- function(X, y, W, b, lambda, n_threads = 4L) {
    .Call(`_FastKRR_rff`, X, y, W, b, lambda, n_threads)
}

get_num_procs <- function() {
    .Call(`_FastKRR_get_num_procs`)
}

solve_chol <- function(A, b) {
    .Call(`_FastKRR_solve_chol`, A, b)
}

SOLVE_sympd <- function(A, b) {
    .Call(`_FastKRR_SOLVE_sympd`, A, b)
}

test_that("make_kernel constructs symmetric kernel matrix when X_new is NULL", {
  set.seed(123)
  n <- 30
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  rho <- 1.0

  # 1. Gaussian kernel symmetric matrix
  K_gauss <- make_kernel(X = X, kernel = "gaussian", rho = rho)
  expect_true(is.matrix(K_gauss))
  expect_equal(dim(K_gauss), c(n, n))
  expect_equal(K_gauss, t(K_gauss)) # Symmetry check
  expect_equal(diag(K_gauss), rep(1.0, n)) # K(x_i, x_i) = exp(0) = 1

  # 2. Laplace kernel symmetric matrix
  K_laplace <- make_kernel(X = X, kernel = "laplace", rho = rho, n_threads = 1)
  expect_true(is.matrix(K_laplace))
  expect_equal(dim(K_laplace), c(n, n))
  expect_equal(K_laplace, t(K_laplace))
  expect_equal(diag(K_laplace), rep(1.0, n))
})

test_that("make_kernel constructs rectangular cross-kernel matrix when X_new is provided", {
  set.seed(123)
  n <- 30
  n_new <- 15
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  X_new <- matrix(runif(n_new * d), nrow = n_new, ncol = d)
  rho <- 1.0

  # 1. Gaussian cross-kernel matrix (n_new x n)
  K_cross_gauss <- make_kernel(X = X, X_new = X_new, kernel = "gaussian", rho = rho)
  expect_true(is.matrix(K_cross_gauss))
  expect_equal(dim(K_cross_gauss), c(n_new, n))

  # 2. Laplace cross-kernel matrix (n_new x n)
  K_cross_laplace <- make_kernel(X = X, X_new = X_new, kernel = "laplace", rho = rho, n_threads = 1)
  expect_true(is.matrix(K_cross_laplace))
  expect_equal(dim(K_cross_laplace), c(n_new, n))
})

utils::globalVariables(c("yhat", "x", "y"))

#' Plot method for fitted Kernel Ridge Regression models
#'
#' @description
#' Visualizes the fitted regression curve from a Kernel Ridge Regression (KRR) model.
#' Automatically generates predictions on a regular grid (1.2 times the training
#' sample size) and overlays them with the training data.
#'
#' @details
#' Currently, \code{plot.krr} supports only uni-variate inputs (\eqn{d = 1}).
#' For multivariate settings (\eqn{d \ge 2}), the plot method will return an error,
#' and users are encouraged to manually slice their data to visualize conditional
#' main effects.
#'
#' @param x A fitted KRR model (class \code{"krr"}) returned by \code{\link{fastkrr}}.
#' @param show_points Logical; if \code{TRUE}, displays the original training data points
#'   as a background layer. Default is \code{TRUE}.
#' @param ... Additional arguments (currently ignored).
#'
#' @return A \code{ggplot} object showing the fitted regression line and training data.
#'
#' @seealso \code{\link{fastkrr}}, \code{\link{predict.krr}}
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' n = 1000
#' rho = 1
#' d = 1
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d); colnames(X) = paste0("X", seq_len(d))
#' y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)
#'
#' data = data.frame(X, y = y)
#'
#' model_exact = fastkrr(data = data, response = "y",
#'                        kernel = "gaussian", rho = rho, opt = "exact", verbose = FALSE)
#' plot(model_exact)
#' }
#'
#' @importFrom ggplot2 ggplot geom_point geom_line labs theme_minimal theme element_text aes
#' @export
plot.krr = function(x, show_points = TRUE, ...) {

  if (!inherits(x, "krr"))
    stop("Input must be a fitted model of class 'krr'.")

  model = x
  X = model$x
  y = as.vector(model$y)
  d = if (is.matrix(X)) ncol(X) else 1
  if (d != 1)
    stop("plot.krr() currently supports only 1D input.")

  n_grid = max(50, round(1.2 * nrow(as.matrix(X))))
  newdata = seq(min(X), max(X), length.out = n_grid)

  pred = predict(model, as.matrix(newdata))

  ord = order(as.numeric(newdata))
  df_pred = data.frame(x = as.numeric(newdata[ord]),
                       yhat = as.numeric(pred[ord]))
  df_train = data.frame(x = as.numeric(X), y = as.numeric(y))

  p = ggplot() +
    {if (show_points)
      geom_point(data = df_train, aes(x = x, y = y),
                 alpha = 0.25, color = "gray")} +
    geom_line(data = df_pred, aes(x = x, y = yhat),
              color = "#0072B2", linewidth = 1) +
    labs(
      title = paste0("FastKRR Fit (opt = '", model$opt, "')"),
      x = "x", y = "Predicted f(x)"
    ) +
    theme_minimal(base_size = 13) +
    theme(plot.title = element_text(hjust = 0.5))

  return(p)
}

#' Compute low-rank approximations (Nyström, Pivoted Cholesky, RFF)
#'
#' Computes low-rank kernel approximation \eqn{\tilde{K} \in \mathbb{R}^{n \times n}} using three methods:
#' Nyström approximation, Pivoted Cholesky decomposition, and
#' Random Fourier Features (RFF).
#'
#' @param X A numeric design matrix \eqn{X \in \mathbb{R}^{n \times d}}.
#' @param opt Method for constructing or approximating:
#'   \describe{
#'     \item{\code{"nystrom"}}{Construct a low-rank approximation of
#'         the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
#'         using the Nyström approximation.}
#'     \item{\code{"pivoted"}}{Construct a low-rank approximation of
#'         the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
#'         using Pivoted Cholesky decomposition.}
#'     \item{\code{"rff"}}{Construct a low-rank approximation of
#'         the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
#'         using Random Fourier Features (RFF).}
#'   }
#' @param kernel Kernel type either "gaussian" or "laplace".
#' @param m Approximation rank (number of random features) for the
#'   low-rank kernel approximation. If not specified, the recommended
#'   choice is
#'   \eqn{\lceil n^{1/2} \cdot \log(d + 5) \rceil}
#'   where \eqn{X} is design matrix, \eqn{n = nrow(X)} and \eqn{d = ncol(X)}.
#' @param rho Scaling parameter of the kernel (\eqn{\rho}), specified by the user.
#' @param eps Tolerance parameter used only in \code{"pivoted"}
#'   for stopping criterion of the Pivoted Cholesky decomposition.
#'   If \code{NULL}, it is dynamically adjusted based on the smoothness of the chosen kernel:
#'   defaults to \code{1e-6} for the infinitely smooth Gaussian kernel to ensure precision,
#'   and \code{1e-4} for the less smooth Laplace kernel.
#' @param W Random frequency matrix \eqn{\omega \in \mathbb{R}^{m \times d}}
#' @param b Random phase vector \eqn{b \in \mathbb{R}^m}, i.i.d. \eqn{\mathrm{Unif} [ 0, 2\pi ]}.
#' @param n_threads Number of parallel threads.
#'   If \code{NULL}, defaults to half of the available system processors.
#'   It automatically falls back to 1 thread if the system has 3 or fewer processors.
#'   Applied only for \code{opt = "nystrom"}, \code{opt = "rff"}, or \code{kernel = "laplace"}.
#'
#' @details
#' Requirements and what to supply:
#'
#' \strong{Common}
#'
#' \itemize{
#'   \item \code{rho} must be provided (non-\code{NULL}).
#' }
#'
#' \strong{nystrom / pivoted}
#'
#' \itemize{
#'   \item If \code{m} is \code{NULL}, use \eqn{\lceil n^{1/2} \cdot \log(d + 5) \rceil}.
#'   \item For \code{"pivoted"}, a tolerance \code{eps} is used; the decomposition stops early
#'   when the next pivot (residual diagonal) drops below \code{eps}.
#' }
#'
#' \strong{rff}
#'
#' \itemize{
#'   \item The function automatically generates
#'         \code{W} (random frequency matrix \eqn{\omega \in \mathbb{R}^{m \times d}}) and
#'         \code{b} (random phase vector \eqn{b \in \mathbb{R}^{m}}).
#'   \item If the user provides them manually, both \code{W} and \code{b} must be specified and their dimensions must be compatible.
#' }
#'
#' @return
#' \itemize{
#'   \item \code{call}: The matched function call used to create the object.
#'   \item \code{opt}: The kernel approximation method actually used (\code{"nystrom", "pivoted", "rff"}).
#'   \item \code{K_approx}: The fully reconstructed \eqn{n \times n} approximated kernel matrix.
#'   \item \code{approx_factor}: Low-rank component matrix (\eqn{R} for Nyström, \eqn{PR} for Pivoted Cholesky, \eqn{Z} for RFF).
#'   \item \code{m}: Kernel approximation degree.
#'   \item \code{rho}: Scaling parameter of the kernel.
#' }
#'
#' Additional components depend on the value of opt:
#'
#' \strong{nystrom}
#' \itemize{
#'   \item \code{n_threads}: Number of threads used in the computation.
#' }
#'
#' \strong{pivoted}
#' \itemize{
#'   \item \code{eps}: Numerical tolerance used for early stopping in the
#'                      pivoted Cholesky decomposition.
#' }
#'
#' \strong{rff}
#' \itemize{
#'   \item \code{d}: Input design matrix's dimension.
#'   \item \code{W}: \eqn{m \times d} Random frequency matrix.
#'   \item \code{b}: Random phase \eqn{m}-vector.
#'   \item \code{used_supplied_Wb}: Logical; \code{TRUE} if user-supplied
#'         \code{W}, \code{b} were used, \code{FALSE} otherwise.
#'   \item \code{n_threads}: Number of threads used in the computation.
#' }
#'
#' @examples
#' # Data setting
#' set.seed(1)
#' d = 1
#' rho = 1
#' n = 100
#' m = 50
#' X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
#'
#' # Example: Nystrom approximation
#' K_nystrom = approx_kernel(X = X, opt = "nystrom",
#'                            m = m, rho = rho, n_threads = 1)
#'
#' # Example: Pivoted Cholesky approximation
#' K_pivoted = approx_kernel(X = X, opt = "pivoted",
#'                            m = m, rho = rho)
#' # Example: RFF approximation
#' K_rff = approx_kernel(X = X, opt = "rff", kernel = "gaussian",
#'                        m = m, rho = rho, n_threads = 1)
#'
#' @export
approx_kernel = function(X = NULL,
                         opt = c("nystrom", "pivoted", "rff"),
                         kernel = c("gaussian", "laplace"),
                         m = NULL, rho, eps = NULL,
                         W = NULL, b = NULL, n_threads = NULL) {
  call   = match.call()
  opt    = match.arg(opt)
  kernel = match.arg(kernel)

  if (is.null(X)) stop("Argument 'X' must be provided (not NULL).", call. = FALSE)
  if (!is.matrix(X)) stop("Argument 'X' must be a numeric matrix.", call. = FALSE)
  d = ncol(X)

  # Adjust eps for opt=="pivoted"
  if (opt == "pivoted") {
    if (is.null(eps)) {
      eps = if (kernel == "gaussian") 1e-6 else 1e-4
    } else {
      if (eps <= 0) stop("eps must be a positive real number", call. = FALSE)
    }
  }

  # Adjust number of threads only for heavy computation cases
  max_threads = get_num_procs()
  if (is.null(n_threads)) n_threads = as.integer(max(1, max_threads %/% 2))
  if ((opt %in% c("nystrom", "rff")) || (kernel == "laplace")){
    n_threads = if (max_threads <= 3) 1 else min(n_threads, max_threads - 1)
  }else{
    n_threads = 1
  }

  if (missing(rho) || is.null(rho)) stop("'rho' must be provided and non-NULL.", call. = FALSE)


  result_values = list()
  class(result_values) = "approx_kernel"
  result_values$call = call
  result_values$opt = opt
  result_values$kernel = kernel

  if (opt %in% c("nystrom", "pivoted")) {

    n = nrow(X)
    if (is.null(m)) m = n^(1/2) * log(d + 5)
    m = as.integer(max(1, min(n, floor(m))))

    if (opt == "nystrom") {
      idx_ny = sample(seq_len(nrow(X)), m)

      K_nm = make_kernel(X[idx_ny, , drop = FALSE], X,
                         kernel = kernel, rho = rho, n_threads = n_threads)

      K_mm = make_kernel(X[idx_ny, , drop = FALSE], X[idx_ny, , drop = FALSE],
                         kernel = kernel, rho = rho, n_threads = n_threads)

      rslt = nystrom_kernel(K_nm = K_nm, K_mm = K_mm, m, n_threads = n_threads)

      result_values$K_approx = tcrossprod(rslt$R)
      result_values$approx_factor = rslt$R
      result_values$m = rslt$m
      result_values$rho = rho
      result_values$n_threads = rslt$n_threads

      return(result_values)
    } else {
      rslt = rslt = pchol_kernel(X, rho = rho, kernel = kernel, m = m, eps = eps)

      result_values$K_approx = tcrossprod(rslt$PR)
      result_values$approx_factor = rslt$PR
      result_values$m = rslt$rank
      result_values$rho = rho
      result_values$eps = rslt$eps
      return(result_values)
    }
  }



  n   = nrow(X)
  d_x = ncol(X)
  if (d != d_x) stop("For opt='rff', provided d must equal ncol(X).")

  if (!is.null(W) || !is.null(b)) {
    if (is.null(W) || is.null(b)) stop("For opt='rff', provide both 'W' and 'b' or neither.")
    if (!is.matrix(W)) stop("'W' must be a numeric matrix (m x d).")
    if (ncol(W) != d) stop("ncol(W) must equal d (and ncol(X)).")
    if (length(b) != nrow(W)) stop("length(b) must equal nrow(W).")

    m_used = nrow(W)
    Z = make_Z(X, W, b, n_threads = n_threads )
    K_approx = tcrossprod(Z)

    result_values$approx_factor = Z
    result_values$K_approx = K_approx
    result_values$m = m_used
    result_values$d = d
    result_values$rho = rho
    result_values$W = W
    result_values$b = b
    result_values$used_supplied_Wb = TRUE
    result_values$n_threads = n_threads

    return(result_values)
  }

  if (is.null(m)) m = n^(1/2) * log(d + 5)
  m = as.integer(max(1, floor(m)))

  old_d_exists = exists("d", envir = .GlobalEnv, inherits = FALSE)
  if (old_d_exists) old_d = get("d", envir = .GlobalEnv)
  on.exit({
  }, add = TRUE)

  rb = rff_random(m = m, rho = rho, d = d, kernel = kernel)
  W = rb$W; b = rb$b
  Z = make_Z(X, W, b, n_threads = n_threads)

  result_values$approx_factor = Z
  result_values$K_approx = tcrossprod(Z)
  result_values$m = m
  result_values$d = d
  result_values$rho = rho
  result_values$W = W
  result_values$b = b
  result_values$used_supplied_Wb = FALSE
  result_values$n_threads = n_threads

  return(result_values)
}

test_that("approx_kernel computes Nystrom approximation correctly", {
  # Set up reproducible synthetic data
  set.seed(123)
  n <- 40
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  m <- 10
  rho <- 1.0

  # Execute Nystrom kernel approximation
  res <- approx_kernel(X = X, opt = "nystrom", kernel = "gaussian", m = m, rho = rho, n_threads = 1)

  # Check object structure, dimensions, and output components
  expect_s3_class(res, "approx_kernel")
  expect_equal(res$opt, "nystrom")
  expect_equal(dim(res$K_approx), c(n, n))
  expect_equal(dim(res$approx_factor), c(n, m))
  expect_equal(res$m, m)
  expect_equal(res$rho, rho)
  expect_false(is.null(res$n_threads))
})

test_that("approx_kernel computes Pivoted Cholesky approximation and handles dynamic eps", {
  set.seed(123)
  n <- 40
  d <- 2
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  m <- 15
  rho <- 1.0

  # Test 1: Gaussian kernel defaults to eps = 1e-6
  res_gauss <- approx_kernel(X = X, opt = "pivoted", kernel = "gaussian", m = m, rho = rho)
  expect_equal(res_gauss$opt, "pivoted")
  expect_equal(dim(res_gauss$K_approx), c(n, n))
  expect_equal(nrow(res_gauss$approx_factor), n)
  expect_equal(res_gauss$eps, 1e-6)

  # Test 2: Laplace kernel defaults to eps = 1e-4
  res_laplace <- approx_kernel(X = X, opt = "pivoted", kernel = "laplace", m = m, rho = rho)
  expect_equal(res_laplace$eps, 1e-4)
})

test_that("approx_kernel handles Random Fourier Features (RFF) with auto & user-supplied W, b", {
  set.seed(123)
  n <- 30
  d <- 3
  X <- matrix(runif(n * d), nrow = n, ncol = d)
  m <- 12
  rho <- 1.0

  # Case 1: Automatically generated W and b
  res_auto <- approx_kernel(X = X, opt = "rff", kernel = "gaussian", m = m, rho = rho, n_threads = 1)
  expect_equal(res_auto$opt, "rff")
  expect_equal(dim(res_auto$K_approx), c(n, n))
  expect_equal(dim(res_auto$approx_factor), c(n, m))
  expect_equal(dim(res_auto$W), c(m, d))
  expect_equal(length(res_auto$b), m)
  expect_false(res_auto$used_supplied_Wb)

  # Case 2: User-supplied W and b
  custom_W <- matrix(rnorm(m * d), nrow = m, ncol = d)
  custom_b <- runif(m, 0, 2 * pi)

  res_supplied <- approx_kernel(X = X, opt = "rff", kernel = "gaussian",
                                rho = rho, W = custom_W, b = custom_b, n_threads = 1)
  expect_true(res_supplied$used_supplied_Wb)
  expect_equal(res_supplied$m, m)
  expect_equal(res_supplied$W, custom_W)
  expect_equal(res_supplied$b, custom_b)
})

test_that("approx_kernel throws expected errors on invalid inputs", {
  set.seed(123)
  X <- matrix(runif(20), nrow = 10, ncol = 2)

  # Error when X is missing
  expect_error(approx_kernel(X = NULL, rho = 1), "Argument 'X' must be provided")

  # Error when rho is missing
  expect_error(approx_kernel(X = X, opt = "nystrom"), "'rho' must be provided")

  # Error when only one of W or b is provided in RFF mode
  custom_W <- matrix(rnorm(10), nrow = 5, ncol = 2)
  expect_error(
    approx_kernel(X = X, opt = "rff", rho = 1, W = custom_W, b = NULL),
    "provide both 'W' and 'b' or neither"
  )
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/error.R
\name{error.krr}
\alias{error.krr}
\alias{error}
\title{Compute Model Error for Kernel Ridge Regression Models}
\usage{
error(x, ...)

\method{error}{krr}(x, data_new = NULL, ...)
}
\arguments{
\item{x}{An object of class \code{"krr"}, typically returned by
\code{\link{fastkrr}}.}

\item{...}{Additional arguments (ignored).}

\item{data_new}{An optional data frame containing new predictor variables
and the observed response variable. The response column must have the
same name as the response variable used to fit the model. If
\code{NULL}, the training MSE is computed. Otherwise, the prediction MSE (PMSE)
is computed using predictions for \code{data_new}.}
}
\value{
A numeric value giving the mean squared error (MSE). If \code{data_new = NULL},
  this returns the training MSE based on the fitted values. If \code{data_new} is provided,
  it returns the prediction mean squared error (PMSE) evaluated on the new dataset.
}
\description{
Computes the model error for kernel ridge regression (\code{"krr"} object).
Returns the mean squared error (MSE) between the observed responses
and the fitted values stored in the object.
}
\details{
This method computes the mean squared error defined as:
\deqn{\text{MSE} = \frac{1}{n} \sum_{i = 1} (y_i - \hat{y}_i)^2}
Depending on the presence of \code{data_new}, the components are defined as follows:
\itemize{
  \item If \code{data_new = NULL}, \eqn{y} is the observed response vector and
    \eqn{\hat{y}} is the fitted values vector, both stored within the \code{"krr"} object.
  \item If \code{data_new} is provided, \eqn{y} is the observed response extracted from
    \code{data_new} and \eqn{\hat{y}} is the predicted values vector generated by
    applying \code{\link{predict.krr}} to the new predictor variables.
}
}
\examples{
# Data setting
set.seed(1)
lambda = 1e-4
d = 1
n = 50
rho = 1
X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)

data = data.frame(X, y = y)

model = fastkrr(data = data, response = "y", kernel = "gaussian",
                 opt = "exact", lambda = lambda)

# MSE
error(model)

new_n = 50
new_x = matrix(runif(new_n*d, 0, 1), nrow = new_n, ncol = d)
new_y = as.vector(sin(2*pi*rowMeans(new_x)^3) + rnorm(new_n, 0, 0.1))
new_data = data.frame(new_x, y = new_y)

# PMSE
error(model, data_new = new_data)

}
\seealso{
\code{\link{summary.krr}}, \code{\link{plot.krr}}, \code{\link{predict.krr}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.krr}
\alias{plot.krr}
\title{Plot method for fitted Kernel Ridge Regression models}
\usage{
\method{plot}{krr}(x, show_points = TRUE, ...)
}
\arguments{
\item{x}{A fitted KRR model (class \code{"krr"}) returned by \code{\link{fastkrr}}.}

\item{show_points}{Logical; if \code{TRUE}, displays the original training data points
as a background layer. Default is \code{TRUE}.}

\item{...}{Additional arguments (currently ignored).}
}
\value{
A \code{ggplot} object showing the fitted regression line and training data.
}
\description{
Visualizes the fitted regression curve from a Kernel Ridge Regression (KRR) model.
Automatically generates predictions on a regular grid (1.2 times the training
sample size) and overlays them with the training data.
}
\details{
Currently, \code{plot.krr} supports only uni-variate inputs (\eqn{d = 1}).
For multivariate settings (\eqn{d \ge 2}), the plot method will return an error,
and users are encouraged to manually slice their data to visualize conditional
main effects.
}
\examples{
\donttest{
set.seed(1)
n = 1000
rho = 1
d = 1
X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d); colnames(X) = paste0("X", seq_len(d))
y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)

data = data.frame(X, y = y)

model_exact = fastkrr(data = data, response = "y",
                       kernel = "gaussian", rho = rho, opt = "exact", verbose = FALSE)
plot(model_exact)
}

}
\seealso{
\code{\link{fastkrr}}, \code{\link{predict.krr}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CVST_linkfunction.R
\name{fastkrr}
\alias{fastkrr}
\title{Fit kernel ridge regression using exact or approximate methods}
\usage{
fastkrr(
  data,
  response,
  kernel = "gaussian",
  opt = "exact",
  m = NULL,
  eps = NULL,
  rho = 1,
  lambda = NULL,
  selection_method = "exactCV",
  n_threads = NULL,
  verbose = TRUE,
  na.rm = FALSE
)
}
\arguments{
\item{data}{A data frame containing the data point variables and response
variable.}

\item{response}{A character string specifying the name of the response
variable in \code{data}.}

\item{kernel}{Kernel type either "gaussian"or "laplace".}

\item{opt}{Method for constructing or approximating :
 \describe{
  \item{\code{"exact"}}{Construct the full kernel matrix
  \eqn{K \in \mathbb{R}^{n\times n}} using design matrix \eqn{X}.}
  \item{\code{"nystrom"}}{Construct a low-rank approximation of
      the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
      using the Nyström approximation.}
  \item{\code{"pivoted"}}{Construct a low-rank approximation of
      the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
      using Pivoted Cholesky decomposition.}
\item{\code{"rff"}}{Use Random Fourier Features to construct a feature map
  \eqn{Z \in \mathbb{R}^{n \times m}} (with \eqn{m} random features) so that
  \eqn{K \approx Z Z^\top}. Here, \eqn{m} is the number of features.}
 }}

\item{m}{Approximation rank(number of random features) used for the low-rank kernel approximation.
If not provided by the user, it defaults to \eqn{\lceil n^{1/2} \cdot \log(d + 5) \rceil},
where \eqn{n = nrow(X)} and \eqn{d = ncol(X)}. Also it must be a positive integer.}

\item{eps}{Tolerance parameter used only in \code{"pivoted"}
for the stopping criterion of the Pivoted Cholesky decomposition.
The default value is dynamically adjusted based on the mathematical smoothness
of the selected kernel: it defaults to \code{1e-6} for the infinitely
differentiable and smooth Gaussian kernel, and relaxes to \code{1e-4} for
the non-smooth, rougher Laplace kernel to properly balance numerical precision
and computational overhead.}

\item{rho}{Scaling parameter of the kernel(\eqn{\rho}),  specified by the user.
  Defaults to \code{1}.
\deqn{\text{Gaussian kernel : } \mathcal{K}(x, x') = \exp(-\rho \| x - x'\|^2_2)}
\deqn{\text{Laplace kernel : } \mathcal{K}(x, x') = \exp(-\rho \| x - x'\|_1)}}

\item{lambda}{Regularization parameter. If \code{NULL}, the penalty parameter
is chosen automatically via \pkg{CVST} package or REML: for \code{selection_method = "REML"}
it defaults to the search range \eqn{[10^{-11}, 10^{-1}]} (a length-2 min/max vector), and for
\code{selection_method \%in\% c("exactCV", "fastCV")} it defaults to a grid of 100 values over
\eqn{[10^{-11}, 10^{-1}]}.}

\item{selection_method}{Method used to select \eqn{\lambda} when a grid or \code{NULL} is
passed. One of:
\describe{
  \item{\code{"exactCV"}}{Full cross-validation via \pkg{CVST} (default).}
  \item{\code{"fastCV"}}{Accelerated sequential-testing CV via \pkg{CVST}.}
  \item{\code{"REML"}}{Restricted Maximum Likelihood.}
}}

\item{n_threads}{Number of parallel threads. If \code{NULL}, it defaults to
half of the available system processors (\code{max_threads \%/\% 2}).
Note that parallelization (implemented in C++) is applied for \code{opt = "nystrom"},
\code{opt = "rff"}, \code{kernel = "laplace"}, or \code{selection_method = "REML"}.
For these parallelizable cases, if the system has 3 or fewer processors, it automatically
falls back to \code{1} thread; otherwise, it is capped at \code{max_threads - 1}.
For all other settings, it is restricted to \code{1} thread.}

\item{verbose}{If TRUE, detailed progress and cross-validation
results are printed to the console. If FALSE, suppresses
intermediate output and only returns the final result.}

\item{na.rm}{Logical. If \code{TRUE}, rows containing missing values are
removed before fitting. Defaults to \code{FALSE}.}
}
\value{
\itemize{
  \item{\code{coefficients}: Estimated coefficient vector. Accessible via \code{model$coefficients}.
    For \code{opt = "rff"}, this is an \eqn{m}-dimensional weight vector in the random feature space.
    For all other options (\code{"exact"}, \code{"pivoted"}, \code{"nystrom"}),
    this is an \eqn{n}-dimensional dual coefficient vector.}
  \item{\code{fitted.values}: Fitted values \eqn{\mathbb{R}^{n}}. Accessible via \code{model$fitted.values}.
    Computed as:
    \itemize{
      \item \eqn{\hat{y} = K \hat{\alpha}} for \code{opt = "exact"} (full kernel matrix \eqn{K}).
      \item \eqn{\hat{y} = \tilde{K} \tilde{\alpha}} for \code{opt = "pivoted"} or \code{"nystrom"} (low-rank kernel matrix \eqn{\tilde{K}}).
      \item \eqn{\hat{y} = Z \hat{\beta}} for \code{opt = "rff"} (random feature matrix \eqn{Z}).
    }
    Here, \eqn{\hat{\alpha}}, \eqn{\tilde{\alpha}}, and \eqn{\hat{\beta}} represent the estimated regression coefficient vectors (\code{coefficients}) obtained under each respective option \code{opt}.}
  \item{\code{opt}: Kernel approximation option.
  One of \code{"exact"}, \code{"pivoted"}, \code{"nystrom"}, \code{"rff"}.}
  \item{\code{kernel}: Kernel used (\code{"gaussian"} or \code{"laplace"}).}
  \item{\code{x}: Input design matrix.}
  \item{\code{y}: Response vector.}
  \item{\code{lambda}: Regularization parameter. If \code{NULL}, tuned
  by cross-validation via \pkg{CVST} or REML.}
  \item{\code{rho}: Additional user-specified hyperparameter.}
  \item{\code{selection_method}: Tunning method for select hyperparmeter lambda}
  \item{\code{call}: The matched function call used to create the object.}
  \item{\code{n_threads}: Number of threads used for parallelization.}
  \item{\code{removed_row_idx}: Indices of rows removed due to missing values.}
}

Additional components depend on the value of \code{opt}:

\subsection{opt = \dQuote{exact}}{
\itemize{
  \item{\code{K}: The full kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}.}
  \item{\code{chol_factor}: Lower triangular Cholesky factor \eqn{L \in \mathbb{R}^{n \times n}}
    of \eqn{K + n\lambda I}; satisfies \eqn{K + n\lambda I = L L^\top}.}
}}


\subsection{opt = \dQuote{nystrom}}{
\itemize{
  \item{\code{m}: Kernel approximation degree.}
  \item{\code{approx_factor}: The method provides a low-rank approximation to the kernel matrix
    \eqn{R \in \mathbb{R}^{n \times m}} obtained via
    Nyström approximation; satisfies \eqn{K \approx R R^\top}.}
}}

\subsection{opt = \dQuote{pivoted}}{
\itemize{
  \item{\code{m}: Kernel pproximation degree.}
  \item{\code{approx_factor}: The method provides a low-rank approximation to the kernel matrix
    \eqn{PR \in \mathbb{R}^{ n \times m}} obtained via
    Pivoted Cholesky decomposition; satisfies \eqn{K \approx PR\,(PR)^\top}.}
  \item{\code{eps}: Numerical tolerance used for early stopping in the pivoted Cholesky decomposition.}
}}

\subsection{opt = \dQuote{rff}}{
\itemize{
  \item{\code{m}: Number of random features.}
  \item{\code{approx_factor}: Random Fourier Feature matrix \eqn{Z \in \mathbb{R}^{n \times m}} with
    \eqn{Z_{ij} = z_j(x_i) = \sqrt{2/m}\cos(\omega_j^\top x_i + b_j), \quad j = 1, \cdots, m,}
    so that \eqn{K \approx Z Z^\top}.}
  \item{\code{W}: Random frequency matrix \eqn{\omega \in \mathbb{R}^{m \times d}}
      (row \eqn{j} is \eqn{\omega_j^\top \in \mathbb{R}^d}), drawn i.i.d. from the spectral density of the chosen kernel:
      \itemize{
        \item {Gaussian: \eqn{\omega_{jk} \sim \mathcal{N}(0, 2\gamma)} (e.g., \eqn{\gamma=1/\ell^2}).}
        \item {Laplace: \eqn{\omega_{jk} \sim \mathrm{Cauchy}(0, 1/\sigma)} i.i.d.
      }}}
  \item{\code{b} Random phase vector \eqn{b \in \mathbb{R}^m}, i.i.d. \eqn{\mathrm{Unif}[0,\,2\pi]}.}
}}
}
\description{
This function performs kernel ridge regression (KRR).
The regularization parameter \eqn{\lambda} can be supplied by
the user or selected automatically using cross-validation or
restricted maximum likelihood (REML). For scalability,
three different kernel approximation strategies are supported (Nyström approximation,
Pivoted Cholesky decomposition, Random Fourier Features(RFF)), and kernel matrix
can be computed using two methods (Gaussian kernel, Laplace kernel).
}
\details{
The function performs several input checks and automatic adjustments:

\itemize{
  \item \code{lambda} can be specified in four ways:
    \enumerate{
      \item A positive numeric scalar, in which case the model is fitted with
         this single value and no selection is performed (any \code{selection_method}).
      \item A numeric vector of length 2 giving \code{c(min, max)}; only valid when
         \code{selection_method = "REML"}, which optimizes \eqn{\lambda} within this range.
      \item A numeric vector (length >= 3) of positive values used as a tuning grid;
         only valid when \code{selection_method \%in\% c("exactCV", "fastCV")}, and
         selection is performed by \pkg{CVST} cross-validation (sequential testing if
         \code{selection_method = "fastCV"}).
      \item \code{NULL}: use a default range/grid (internal setting) and tune \code{lambda}
        via \pkg{CVST} or REML, depending on \code{selection_method}.}
}

After fitting the model with \code{fastkrr()}, users can readily generate predictions
for new test data or out-of-sample observations using the standard generic
\code{\link[stats]{predict}} function, which is internally dispatched to \code{\link{predict.krr}}.
}
\examples{
# Data setting
set.seed(1)
lambda = 1e-4
d = 3
rho = 1
n = 50

X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)

data = data.frame(X, y = y)

# Example: pivoted cholesky
model = fastkrr(data = data, response = "y",  kernel = "gaussian",
                opt = "pivoted", rho = rho, lambda = lambda)

# Example: nystrom
model = fastkrr(data = data, response = "y", kernel = "gaussian",
                opt = "nystrom", rho = rho, lambda = lambda)

# Example: random fourier features
model = fastkrr(data = data, response = "y", kernel = "gaussian",
                opt = "rff", rho = rho, lambda = lambda)

# Example: Laplace kernel
model = fastkrr(data = data, response = "y",  kernel = "laplace",
                opt = "nystrom", n_threads = 1, rho = rho)

# Generate predictions for new data
new_X = matrix(runif(10 * d, 0, 1), nrow = 10, ncol = d)
pred = predict(model, new_X)
print(pred)

}
\seealso{
\code{\link{predict.krr}}, \code{\link{param.krr}}, \code{\link{make_kernel}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FastKRR-package.R
\docType{package}
\name{FastKRR-package}
\alias{FastKRR-package}
\alias{FastKRR}
\title{Kernel Ridge Regression using the \pkg{RcppArmadillo} Package}
\description{
The \pkg{FastKRR} implements its core computational operations in C++ via \pkg{RcppArmadillo}, enabling faster performance than pure R, improved numerical stability, and parallel execution with OpenMP where available. On systems without OpenMP support, the package automatically falls back to single-threaded execution with no user configuration required. For efficient model selection, it integrates with \pkg{CVST} to provide full and sequential-testing cross-validation and additionally supports restricted maximum likelihood (REML) for continuous optimization of the regularization parameter. The package offers a unified interface for exact kernel ridge regression and three widely used scalable approximations—Nyström, Pivoted Cholesky, and Random Fourier Features—allowing analyses with substantially larger sample sizes than are feasible with exact KRR while retaining strong predictive performance. This combination of a compiled backend and scalable algorithms addresses limitations of packages that rely solely on exact computation, which is often impractical for large n. It also integrates with the \pkg{tidymodels} ecosystem via the \pkg{parsnip} model specification \code{krr_reg}, and the S3 method \code{tunable.krr_reg()} (exposes tunable parameters to \code{dials}/\code{tune}); see their help pages for usage.
}
\section{Directory structure}{

\itemize{
  \item \code{R/}: High-level R functions and user-facing API
  \item \code{src/}: C++ sources (kernel computation, fitting, prediction)
}
This package links against \CRANpkg{Rcpp} and \CRANpkg{RcppArmadillo}
(via LinkingTo). It uses \CRANpkg{CVST}, \CRANpkg{parsnip}, and the
\CRANpkg{tidymodels} ecosystem through their public R APIs.
}

\seealso{
\CRANpkg{CVST}, \CRANpkg{Rcpp}, \CRANpkg{RcppArmadillo}, \CRANpkg{parsnip}, \CRANpkg{tidymodels}
}
\author{
\strong{Maintainer}: Kwan-Young Bak \email{kybak@sungshin.ac.kr} (\href{https://orcid.org/0000-0002-4541-160X}{ORCID}) (Sungshin Women's University) [copyright holder]

Authors:
\itemize{
  \item Gyeongmin Kim \email{rlarudals0824@gmail.com} (Sungshin Women's University)
  \item Seyoung Lee \email{sudang0404@gmail.com} (Sungshin Women's University)
  \item Miyoung Jang \email{miyoung9072@gmail.com} (Sungshin Women's University)
}

}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_kernel.R
\name{make_kernel}
\alias{make_kernel}
\title{Kernel matrix construction for given datasets}
\usage{
make_kernel(X,
            X_new = NULL,
            kernel = "gaussian",
            rho = 1,
            n_threads = NULL)
}
\arguments{
\item{X}{Design matrix \eqn{X \in \mathbb{R}^{n \times d}} (rows  \eqn{x_i \in \mathbb{R}^d}).}

\item{X_new}{Second matrix \eqn{X' \in \mathbb{R}^{n' \times d}} (rows  \eqn{x'_j \in \mathbb{R}^d}).
If omitted, \eqn{X' = X} and \eqn{n' = n}.}

\item{kernel}{Kernel type; one of \code{"gaussian"} or \code{"laplace"}.}

\item{rho}{Kernel width parameter (\eqn{\rho > 0}). Default is 1.}

\item{n_threads}{Number of parallel threads. If \code{NULL}, it defaults to
half of the available system processors (\code{max_threads \%/\% 2}).
For these parallelizable cases, if the system has 3 or fewer processors, it automatically
falls back to \code{1} thread; otherwise, it is capped at \code{max_threads - 1}.
For all other settings, it is restricted to \code{1} thread.
Parallelization (implemented in C++) is one of the main advantages
of this package and is applied only for \code{"laplace"} kernels.}
}
\value{
The computed kernel matrix. If \code{X_new} is \code{NULL}, the result is a symmetric matrix
\eqn{K_{ij} = \mathcal{K}(x_i, x_j)}, with \eqn{K \in \mathbb{R}^{n \times n}}.
Otherwise, the result is a rectangular matrix
\eqn{K'_{ji} = \mathcal{K}(x_i, x'_j)}, with \eqn{K' \in \mathbb{R}^{n' \times n}}.
}
\description{
Constructs a kernel matrix \eqn{K \in \mathbb{R}^{n' \times n}} given two
datasets \eqn{X \in \mathbb{R}^{n \times d}} and \eqn{X' \in \mathbb{R}^{n' \times d}},
where \eqn{x_i \in \mathbb{R}^d} and \eqn{x'_j \in \mathbb{R}^d} denote the
i-th and j-th rows of \eqn{X} and \eqn{X'}, respectively, and
\eqn{K_{ji}=\mathcal{K}(x_i, x'_j)} for a user-specified kernel.
Implemented in C++ via RcppArmadillo.
}
\details{
Gaussian kernel: \deqn{K_{ji}=\mathcal{K}(x_i,x_j)=\exp\!\big(-\rho\|x_i-x_j\|_2^2\big)}
Laplace kernel:  \deqn{K_{ji}=\mathcal{K}(x_i,x_j)=\exp\!\big(-\rho\|x_i-x_j\|_1\big)}
}
\examples{
# Data setting
set.seed(1)
d = 1
rho = 1
n = 100
X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)

# New design matrix
new_n = 150
new_X = matrix(runif(new_n*d, 0, 1), nrow = new_n, ncol = d)

# Make kernel : Gaussian kernel
K = make_kernel(X, kernel = "gaussian", rho = rho) ## symmetric matrix
new_K = make_kernel(X, new_X, kernel = "gaussian", rho = rho) ## rectangular matrix

# Make kernel : Laplace kernel
K = make_kernel(X, kernel = "laplace", rho = rho, n_threads = 1) ## symmetric matrix
new_K = make_kernel(X, new_X, kernel = "laplace", rho = rho, n_threads = 1) ## rectangular matrix

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/link2tidymodels.R
\name{tunable.krr_reg}
\alias{tunable.krr_reg}
\title{Expose tunable parameters for \code{"krr_reg"}}
\usage{
\method{tunable}{krr_reg}(x, ...)
}
\arguments{
\item{x}{A \code{"krr_reg"} model specification.}

\item{...}{Not used; included for S3 method compatibility.}
}
\value{
A tibble (one row per tunable parameter)
   with columns \code{"name"}, \code{"call_info"}, \code{"source"},
  \code{"component"}, and \code{"component_id"}.
}
\description{
Supplies a tibble of tunable arguments for \code{"krr_reg()"}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CVST_linkfunction.R
\name{predict.krr}
\alias{predict.krr}
\title{Predict responses for new data using fitted KRR model}
\usage{
\method{predict}{krr}(object, newdata, ...)
}
\arguments{
\item{object}{A S3 object of class \code{krr} created by \code{\link{fastkrr}}.}

\item{newdata}{A numeric design matrix containing new observations
for which predictions are to be made. If \code{newdata} is missing, the function returns fitted values.}

\item{...}{Additional arguments (currently ignored).}
}
\value{
A numeric vector of predicted values corresponding to \code{newdata} or fitted values.
}
\description{
Generates predictions from a fitted Kernel Ridge Regression (KRR) model
for new data.
}
\examples{
# Data setting
n = 30
d = 1

X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)

data = data.frame(X, y = y)

lambda = 1e-4
rho = 1

# Fitting model: pivoted
model = fastkrr(data = data, response = "y",
                kernel = "gaussian", rho = rho, lambda = lambda, opt = "pivoted")

# Predict
new_n = 50
new_x = matrix(runif(new_n*d, 0, 1), nrow = new_n, ncol = d)
new_y = as.vector(sin(2*pi*rowMeans(new_x)^3) + rnorm(new_n, 0, 0.1))

pred = predict(model, new_x)
crossprod(pred - new_y) / new_n

}
\seealso{
\code{\link{fastkrr}}, \code{\link{make_kernel}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coef.R
\name{coef.krr}
\alias{coef.krr}
\title{Extract Model Coefficients from a Fitted KRR Model}
\usage{
\method{coef}{krr}(object, ...)
}
\arguments{
\item{object}{An S3 object of class \code{krr}, typically returned by
\code{\link{fastkrr}}.}

\item{...}{Additional arguments (currently ignored).}
}
\value{
A numeric vector of the estimated model coefficients (\eqn{\alpha} or \eqn{\beta}).
}
\description{
Extracts the estimated coefficients from a fitted Kernel Ridge Regression (KRR) model.
The type of coefficient reported depends on the kernel approximation method:
for \code{opt = "exact"}, \code{"nystrom"}, or \code{"pivoted"}, the coefficients represent
the dual weights \eqn{\alpha}. For \code{opt = "rff"}, they represent the coefficients \eqn{\beta}.
}
\examples{
# Data setting
set.seed(1)
lambda = 1e-4
d = 1
n = 50
rho = 1
X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)

data = data.frame(X, y = y)

# Example: exact
model = fastkrr(data = data, response = "y",
                 kernel = "gaussian", opt = "exact",
                 rho = rho, lambda = lambda)

coef(model)

}
\seealso{
\code{\link{fastkrr}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/link2tidymodels.R
\name{fastkrr_fit_wrapper}
\alias{fastkrr_fit_wrapper}
\title{Internal bridge for parsnip fit interface}
\usage{
fastkrr_fit_wrapper(
  x,
  y,
  kernel = "gaussian",
  opt = "exact",
  m = NULL,
  eps = 1e-06,
  rho = 1,
  lambda = NULL,
  n_threads = 4,
  verbose = FALSE,
  na.rm = FALSE
)
}
\description{
Internal bridge for parsnip fit interface
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/link2tidymodels.R
\name{krr_reg}
\alias{krr_reg}
\title{Kernel Ridge Regression}
\usage{
krr_reg(
  mode = "regression",
  kernel = NULL,
  opt = NULL,
  eps = NULL,
  n_threads = NULL,
  m = NULL,
  rho = NULL,
  penalty = NULL,
  na.rm = NULL
)
}
\arguments{
\item{mode}{A single string; only `"regression"` is supported.}

\item{kernel}{Kernel matrix has two kinds of Kernel ("gaussian", "laplace").}

\item{opt}{Method for constructing or approximating :
 \describe{
  \item{\code{"exact"}}{Construct the full kernel matrix
  \eqn{K \in \mathbb{R}^{n\times n}} using design matrix \eqn{X}.}
  \item{\code{"nystrom"}}{Construct a low-rank approximation of
      the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
      using the Nyström approximation.}
  \item{\code{"pivoted"}}{Construct a low-rank approximation of
      the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
      using Pivoted Cholesky decomposition.}
\item{\code{"rff"}}{Use Random Fourier Features to construct a feature map
  \eqn{Z \in \mathbb{R}^{n \times m}} (with \eqn{m} random features) so that
  \eqn{K \approx Z Z^\top}. Here, \eqn{m} is the number of features.}
 }}

\item{eps}{Tolerance parameter used only in \code{"pivoted"}
for stopping criterion of the Pivoted Cholesky decomposition.}

\item{n_threads}{Number of parallel threads. It is applied only for
\code{opt = "nystrom"} or \code{opt = "rff"}, and for the
 Laplace kernel (\code{kernel = "laplace"}).}

\item{m}{Approximation rank(number of random features) used for the low-rank kernel approximation.}

\item{rho}{Scaling parameter of the kernel(\eqn{\rho}).}

\item{penalty}{Regularization parameter. It must be a positive numeric
value, a numeric vector containing at least three positive values, or
\code{tune()}. When using the \pkg{tidymodels} interface for
hyperparameter selection, setting \code{penalty = tune()} is recommended.
Candidate values are then supplied through the \code{grid} argument of
\code{tune_grid()} and evaluated using the resampling scheme specified
by the user.}

\item{na.rm}{Logical. If \code{TRUE}, rows containing missing values are
removed before fitting. Defaults to \code{FALSE}.}
}
\value{
A parsnip model specification of class \code{"krr_reg"}.
}
\description{
Defines a Kernel Ridge Regression model specification
for use with the tidymodels ecosystem via \pkg{parsnip}. This spec can be
paired with the \code{"fastkrr"} engine implemented in this package to fit
exact or kernel approximation (Nyström, Pivoted Cholesky, Random Fourier Features) within
\pkg{recipes}/\pkg{workflows} pipelines.
}
\examples{
\donttest{
if (all(vapply(
  c("parsnip","stats","modeldata"),
  requireNamespace, quietly = TRUE, FUN.VALUE = logical(1)
))) {
library(tidymodels)
library(parsnip)
library(stats)
library(modeldata)

# Data analysis
data(ames)
ames = ames \%>\% mutate(Sale_Price = log10(Sale_Price))

set.seed(502)
ames_split = initial_split(ames, prop = 0.80, strata = Sale_Price)
ames_train = training(ames_split) # dim (2342, 74)
ames_test  = testing(ames_split) # dim (588, 74)

# Model spec
krr_spec = krr_reg(kernel = "gaussian", opt = "nystrom",
                   m = 50, eps = 1e-6, n_threads = 1,
                   rho = 1, penalty = tune()) \%>\%
 set_engine("fastkrr") \%>\%
 set_mode("regression")

# Define rec
rec = recipe(Sale_Price ~ Longitude + Latitude, data = ames_train)

# workflow
wf = workflow() \%>\%
  add_recipe(rec) \%>\%
  add_model(krr_spec)

# Define hyper-parameter grid
param_grid = grid_regular(
  dials::penalty(range = c(-10, -3)),
  levels = 5
)

# CV setting
set.seed(123)
cv_folds = vfold_cv(ames_train, v = 5, strata = Sale_Price)

# Tuning
tune_results = tune_grid(
  wf,
  resamples = cv_folds,
  grid = param_grid,
  metrics = metric_set(rmse),
  control = control_grid(verbose = TRUE, save_pred = TRUE)
)

# Result check
collect_metrics(tune_results)

# Select best parameter
best_params = select_best(tune_results, metric = "rmse")

# Finalized model spec using best parameter
final_spec = finalize_model(krr_spec, best_params)
final_wf = workflow() \%>\%
  add_recipe(rec) \%>\%
  add_model(final_spec)

# Finalized fitting using best parameter
final_fit = final_wf \%>\% fit(data = ames_train)

# Prediction
predict(final_fit, new_data = ames_test)
print(best_params)

}}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.approx_kernel}
\alias{print.approx_kernel}
\title{Print method for approximated kernel matrices}
\usage{
\method{print}{approx_kernel}(x, ...)
}
\arguments{
\item{x}{An S3 object created by \code{\link{approx_kernel}}.}

\item{...}{Additional arguments (currently ignored).}
}
\value{
Invisibly returns the input \code{approx_kernel} object after printing the options.
}
\description{
Displays the key algorithmic choices and hyperparameter options used to construct
an approximated kernel matrix object.
}
\details{
The function summarizes critical metadata attributes stored within the
\code{approx_kernel} object, including the approximation strategy (\code{opt}),
kernel type, approximation degree (\code{m}), numerical tolerance (\code{eps}),
and the number of parallel computational threads used.
}
\examples{
# Data setting
set.seed(1)
d = 1
n = 100
rho = 1
X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)

# Example: nystrom
K_nystrom = approx_kernel(X = X, opt = "nystrom", rho = rho, n_threads = 1)

print(K_nystrom)
}
\seealso{
\code{\link{approx_kernel}}, \code{\link{print.krr}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/param.R
\name{param.krr}
\alias{param.krr}
\title{Displays hyperparameters of fitted Kernel Ridge Regression models}
\usage{
\method{param}{krr}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"krr"}, typically returned by
\code{\link{fastkrr}}.}

\item{...}{Additional arguments.}
}
\value{
Prints a human-readable panel to the console and invisibly returns a
named list of class \code{"krr_params"} containing the extracted hyperparameters:
\itemize{
  \item \code{kernel}: Kernel type used ("gaussian" or "laplace").
  \item \code{opt}: Kernel approximation method.
  \item \code{selection_method}: Lambda tuning method.
  \item \code{rho}: Kernel scaling parameter.
  \item \code{lambda}: Regularization parameter.
  \item \code{m}: Rank or number of random features used for approximation.
  \item \code{eps}: Tolerance parameter (for pivoted Cholesky).
  \item \code{n_threads}: Number of parallel threads used.
}
}
\description{
Displays (and invisibly returns) the hyperparameters actually used by a
fitted object. For \code{krr} objects returned by \code{\link{fastkrr}},
this prints a concise hyperparameter panel (e.g., \code{rho}, \code{lambda},
\code{m}, \code{eps}, \code{n_threads}, \code{d}).
}
\details{
\strong{Pivoted approximation note}: When \code{opt = "pivoted"}, the
effective number of pivots \code{m} used during the approximation may be
smaller than the user-specified \code{m} because the algorithm can stop
early based on \code{eps}. If you want to confirm the initial \code{m}
that you set, please see the printed \emph{Call} (the original function
call shows your input arguments).
}
\examples{
# Data setting
set.seed(1)
lambda = 1e-4
d = 1
n = 50
rho = 1
X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)

data = data.frame(X, y = y)

model = fastkrr(data = data, response = "y",
                 kernel="gaussian", opt="pivoted",
                 rho=1, lambda=lambda, n_threads = 1)

# Inspect hyperparameters
param(model)

}
\seealso{
\code{\link{fastkrr}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/approx_kernel.R
\name{approx_kernel}
\alias{approx_kernel}
\title{Compute low-rank approximations (Nyström, Pivoted Cholesky, RFF)}
\usage{
approx_kernel(
  X = NULL,
  opt = c("nystrom", "pivoted", "rff"),
  kernel = c("gaussian", "laplace"),
  m = NULL,
  rho,
  eps = NULL,
  W = NULL,
  b = NULL,
  n_threads = NULL
)
}
\arguments{
\item{X}{A numeric design matrix \eqn{X \in \mathbb{R}^{n \times d}}.}

\item{opt}{Method for constructing or approximating:
\describe{
  \item{\code{"nystrom"}}{Construct a low-rank approximation of
      the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
      using the Nyström approximation.}
  \item{\code{"pivoted"}}{Construct a low-rank approximation of
      the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
      using Pivoted Cholesky decomposition.}
  \item{\code{"rff"}}{Construct a low-rank approximation of
      the kernel matrix \eqn{K \in \mathbb{R}^{n \times n}}
      using Random Fourier Features (RFF).}
}}

\item{kernel}{Kernel type either "gaussian" or "laplace".}

\item{m}{Approximation rank (number of random features) for the
low-rank kernel approximation. If not specified, the recommended
choice is
\eqn{\lceil n^{1/2} \cdot \log(d + 5) \rceil}
where \eqn{X} is design matrix, \eqn{n = nrow(X)} and \eqn{d = ncol(X)}.}

\item{rho}{Scaling parameter of the kernel (\eqn{\rho}), specified by the user.}

\item{eps}{Tolerance parameter used only in \code{"pivoted"}
for stopping criterion of the Pivoted Cholesky decomposition.
If \code{NULL}, it is dynamically adjusted based on the smoothness of the chosen kernel:
defaults to \code{1e-6} for the infinitely smooth Gaussian kernel to ensure precision,
and \code{1e-4} for the less smooth Laplace kernel.}

\item{W}{Random frequency matrix \eqn{\omega \in \mathbb{R}^{m \times d}}}

\item{b}{Random phase vector \eqn{b \in \mathbb{R}^m}, i.i.d. \eqn{\mathrm{Unif} [ 0, 2\pi ]}.}

\item{n_threads}{Number of parallel threads.
If \code{NULL}, defaults to half of the available system processors.
It automatically falls back to 1 thread if the system has 3 or fewer processors.
Applied only for \code{opt = "nystrom"}, \code{opt = "rff"}, or \code{kernel = "laplace"}.}
}
\value{
\itemize{
  \item \code{call}: The matched function call used to create the object.
  \item \code{opt}: The kernel approximation method actually used (\code{"nystrom", "pivoted", "rff"}).
  \item \code{K_approx}: The fully reconstructed \eqn{n \times n} approximated kernel matrix.
  \item \code{approx_factor}: Low-rank component matrix (\eqn{R} for Nyström, \eqn{PR} for Pivoted Cholesky, \eqn{Z} for RFF).
  \item \code{m}: Kernel approximation degree.
  \item \code{rho}: Scaling parameter of the kernel.
}

Additional components depend on the value of opt:

\strong{nystrom}
\itemize{
  \item \code{n_threads}: Number of threads used in the computation.
}

\strong{pivoted}
\itemize{
  \item \code{eps}: Numerical tolerance used for early stopping in the
                     pivoted Cholesky decomposition.
}

\strong{rff}
\itemize{
  \item \code{d}: Input design matrix's dimension.
  \item \code{W}: \eqn{m \times d} Random frequency matrix.
  \item \code{b}: Random phase \eqn{m}-vector.
  \item \code{used_supplied_Wb}: Logical; \code{TRUE} if user-supplied
        \code{W}, \code{b} were used, \code{FALSE} otherwise.
  \item \code{n_threads}: Number of threads used in the computation.
}
}
\description{
Computes low-rank kernel approximation \eqn{\tilde{K} \in \mathbb{R}^{n \times n}} using three methods:
Nyström approximation, Pivoted Cholesky decomposition, and
Random Fourier Features (RFF).
}
\details{
Requirements and what to supply:

\strong{Common}

\itemize{
  \item \code{rho} must be provided (non-\code{NULL}).
}

\strong{nystrom / pivoted}

\itemize{
  \item If \code{m} is \code{NULL}, use \eqn{\lceil n^{1/2} \cdot \log(d + 5) \rceil}.
  \item For \code{"pivoted"}, a tolerance \code{eps} is used; the decomposition stops early
  when the next pivot (residual diagonal) drops below \code{eps}.
}

\strong{rff}

\itemize{
  \item The function automatically generates
        \code{W} (random frequency matrix \eqn{\omega \in \mathbb{R}^{m \times d}}) and
        \code{b} (random phase vector \eqn{b \in \mathbb{R}^{m}}).
  \item If the user provides them manually, both \code{W} and \code{b} must be specified and their dimensions must be compatible.
}
}
\examples{
# Data setting
set.seed(1)
d = 1
rho = 1
n = 100
m = 50
X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)

# Example: Nystrom approximation
K_nystrom = approx_kernel(X = X, opt = "nystrom",
                           m = m, rho = rho, n_threads = 1)

# Example: Pivoted Cholesky approximation
K_pivoted = approx_kernel(X = X, opt = "pivoted",
                           m = m, rho = rho)
# Example: RFF approximation
K_rff = approx_kernel(X = X, opt = "rff", kernel = "gaussian",
                       m = m, rho = rho, n_threads = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/param.R
\name{param}
\alias{param}
\alias{param.default}
\title{Extract/print hyperparameters of fitted models}
\usage{
param(x, ...)

\method{param}{default}(x, ...)
}
\arguments{
\item{x}{An object.}

\item{...}{Additional arguments passed to methods.}
}
\value{
A named list of hyperparameters (invisibly); may print side effects.
}
\description{
\code{param} is a generic S3 function that displays (and invisibly returns)
model hyperparameters. Methods are provided for \code{"krr"} objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.krr}
\alias{summary.krr}
\title{Summary method for fitted Kernel Ridge Regression models}
\usage{
\method{summary}{krr}(object, ...)
}
\arguments{
\item{object}{An S3 object of class \code{krr}, typically returned by
\code{\link{fastkrr}}.}

\item{...}{Additional arguments (currently ignored).}
}
\value{
Invisibly returns the hyperparameter list or model statistics after
  printing the summary panel to the console.
}
\description{
Computes and displays a comprehensive summary of a fitted Kernel Ridge Regression (KRR) model,
including the original function call, fitted hyperparameters (via \code{\link{param.krr}}),
and the final training mean squared error (via \code{\link{error.krr}}).
}
\examples{
# Data setting
set.seed(1)
lambda = 1e-4
d = 1
n = 50
rho = 1
X = matrix(runif(n*d, 0, 1), nrow = n, ncol = d)
y = sin(2 * pi * rowMeans(X)^3) + rnorm(n, mean = 0, sd = 0.1)

data = data.frame(X, y = y)

# Example: exact
model = fastkrr(data = data, response = "y",
                 kernel = "gaussian", opt = "exact",
                 rho = rho, lambda = lambda)

summary(model)

}
\seealso{
\code{\link{fastkrr}}, \code{\link{param.krr}}, \code{\link{error.krr}}
}
